from typing import Callable, Mapping, Optional, NamedTuple

from pyasn1.type import univ
from pyasn1_alt_modules import rfc5280, rfc6962, rfc6960

from pkilint import validation, document, pkix, oid
from pkilint.cabf.asn1 import ev_guidelines
from pkilint.itu import x520_name
from pkilint.pkix import Rfc2119Word

OID_TO_CODE_NAME = {
    # EKU
    rfc5280.id_kp_codeSigning: 'codesigning',
    rfc5280.id_kp_emailProtection: 'emailprotection',
    rfc5280.id_kp_timeStamping: 'timestamping',
    rfc5280.id_kp_OCSPSigning: 'ocspsigning',
    rfc5280.anyExtendedKeyUsage: 'anyeku',
    rfc6962.id_kp_precertificateSigning: 'precertsigning',
    rfc5280.id_kp_serverAuth: 'serverauth',
    rfc5280.id_kp_clientAuth: 'clientauth',

    # extension
    rfc5280.id_ce_authorityKeyIdentifier: 'authority_key_identifier',
    rfc5280.id_ce_basicConstraints: 'basic_constraints',
    rfc5280.id_ce_certificatePolicies: 'certificate_policies',
    rfc5280.id_ce_cRLDistributionPoints: 'crl_distribution_points',
    rfc5280.id_ce_keyUsage: 'key_usage',
    rfc5280.id_ce_subjectKeyIdentifier: 'subject_key_identifier',
    rfc5280.id_ce_extKeyUsage: 'extended_key_usage',
    rfc5280.id_pe_authorityInfoAccess: 'authority_info_access',
    rfc5280.id_ce_nameConstraints: 'name_constraints',
    rfc6962.id_ce_embeddedSCT: 'sct_list',
    rfc5280.id_ce_subjectAltName: 'subject_altname',
    rfc6962.id_ce_criticalPoison: 'precert_poison',
    rfc6960.id_pkix_ocsp_nocheck: 'ocsp_nocheck',

    # AIA access methods
    rfc5280.id_ad_ocsp: 'ocsp',
    rfc5280.id_ad_caIssuers: 'ca_issuers',

    # attributes
    rfc5280.id_at_countryName: 'country',
    rfc5280.id_at_stateOrProvinceName: 'state_or_province',
    rfc5280.id_at_localityName: 'locality',
    x520_name.id_at_postalCode: 'postal_code',
    x520_name.id_at_streetAddress: 'street_address',
    rfc5280.id_at_organizationName: 'organization_name',
    rfc5280.id_at_commonName: 'common_name',
    rfc5280.id_at_organizationalUnitName: 'organizational_unit_name',
    rfc5280.id_at_serialNumber: 'serial_number',
    x520_name.id_at_businessCategory: 'business_category',
    ev_guidelines.id_evat_jurisdiction_countryName: 'jurisdiction_country',
    ev_guidelines.id_evat_jurisdiction_stateOrProvinceName: 'jurisdiction_state_or_province',
    ev_guidelines.id_evat_jurisdiction_localityName: 'jurisdiction_locality',
    rfc5280.id_at_surname: 'surname',
    rfc5280.id_at_givenName: 'given_name',
}


class ElementIdentifierAllowanceValidator(validation.Validator):
    @staticmethod
    def _create_finding(fmt: str, rfc2119word: pkix.Rfc2119Word, o: univ.ObjectIdentifier):
        if rfc2119word == pkix.Rfc2119Word.MAY:
            return None
        else:
            return validation.ValidationFinding(rfc2119word.to_severity, fmt.format(oid=OID_TO_CODE_NAME[o]))

    def __init__(self, element_name: str, element_oid_retriever: Callable[[document.PDUNode], document.PDUNode],
                 known_element_allowances: Mapping[univ.ObjectIdentifier, pkix.Rfc2119Word],
                 unexpected_presence_code_format: str = None,
                 unexpected_absence_code_format: str = None,
                 unknown_element_presence_finding: Optional[validation.ValidationFinding] = None,
                 **kwargs
                 ):
        self._element_name = element_name
        self._element_oid_retriever = element_oid_retriever

        self._expected_element_presences = {
            o: ElementIdentifierAllowanceValidator._create_finding(unexpected_absence_code_format, w, o)
            for o, w in known_element_allowances.items()
            if w in {Rfc2119Word.MAY, Rfc2119Word.SHOULD, Rfc2119Word.MUST}
        }
        self._expected_element_absences = {
            o: ElementIdentifierAllowanceValidator._create_finding(unexpected_presence_code_format, w, o)
            for o, w in known_element_allowances.items()
            if w in {Rfc2119Word.MAY, Rfc2119Word.SHOULD_NOT, Rfc2119Word.MUST_NOT}
        }

        self._unknown_element_presence_finding = unknown_element_presence_finding

        validations = (
                [a for a in self._expected_element_presences.values() if a is not None] +
                [a for a in self._expected_element_absences.values() if a is not None]
        )

        if unknown_element_presence_finding is not None:
            validations.append(unknown_element_presence_finding)

        super().__init__(validations=list(validations), **kwargs)

    def validate(self, node):
        oids = {self._element_oid_retriever(n).pdu for n in node.children.values()}

        # process unexpected present elements
        finding_descriptions = [validation.ValidationFindingDescription(v, None)
                                for o, v in self._expected_element_absences.items()
                                if o in oids and v is not None]

        # process unexpected absent elements
        finding_descriptions.extend([validation.ValidationFindingDescription(v, None)
                                     for o, v in self._expected_element_presences.items()
                                     if o not in oids and v is not None])

        # process unknown elements
        if self._unknown_element_presence_finding is not None:
            finding_descriptions.extend([
                validation.ValidationFindingDescription(self._unknown_element_presence_finding,
                                                        f'Unknown {self._element_name} present: {str(o)}')
                for o in oids if o not in self._expected_element_presences and o not in self._expected_element_absences
            ])

        return validation.ValidationResult(self, node, finding_descriptions)


class ExtensionsPresenceValidator(validation.Validator):
    def __init__(self, validation):
        self._validation = validation

        super().__init__(validations=validation, pdu_class=rfc5280.TBSCertificate)

    def validate(self, node):
        if 'extensions' not in node.children:
            raise validation.ValidationFindingEncountered(self._validation)


class ExtensionIdentifierAllowanceValidator(ElementIdentifierAllowanceValidator):
    @staticmethod
    def _retrieve_extension_id(node):
        return node.children['extnID']

    def __init__(self, extension_allowances, finding_code_classifier: str, unknown_extension_allowance: Rfc2119Word):
        unknown_extension_finding = None if unknown_extension_allowance == Rfc2119Word.MAY else (
            validation.ValidationFinding(
                unknown_extension_allowance.to_severity,
                f'{finding_code_classifier}.unknown_extension_present'
            )
        )

        super().__init__('extension', ExtensionIdentifierAllowanceValidator._retrieve_extension_id,
                         extension_allowances,
                         f'{finding_code_classifier}.{{oid}}_extension_present',
                         f'{finding_code_classifier}.{{oid}}_extension_absent',
                         unknown_extension_finding, pdu_class=rfc5280.Extensions)


class ExtendedKeyUsageAllowanceValidator(ElementIdentifierAllowanceValidator):
    @staticmethod
    def _retrieve_eku(node):
        return node

    def __init__(self, eku_allowances, finding_code_classifier: str, unknown_eku_allowance: Rfc2119Word):
        unknown_eku_finding = None if unknown_eku_allowance == Rfc2119Word.MAY else (
            validation.ValidationFinding(
                unknown_eku_allowance.to_severity,
                f'{finding_code_classifier}.unknown_eku_present'
            )
        )
        super().__init__('EKU', ExtendedKeyUsageAllowanceValidator._retrieve_eku, eku_allowances,
                         f'{finding_code_classifier}.{{oid}}_eku_present',
                         f'{finding_code_classifier}.{{oid}}_eku_absent',
                         unknown_eku_finding, pdu_class=rfc5280.ExtKeyUsageSyntax)


class ExtensionCriticalityValidator(validation.Validator):
    def __init__(self, criticality_mapping: Mapping[univ.ObjectIdentifier, bool], finding_code_classifier: str,
                 critical_adherence_word: Rfc2119Word, non_critical_adherence_word: Rfc2119Word):
        self._expected_critical_extensions = {
            o: validation.ValidationFinding(
                critical_adherence_word.to_severity,
                f'{finding_code_classifier}.non_critical_{OID_TO_CODE_NAME[o]}_extension'
            ) for o, c in criticality_mapping.items() if c and critical_adherence_word is not Rfc2119Word.MAY
        }
        self._expected_non_critical_extensions = {
            o: validation.ValidationFinding(
                critical_adherence_word.to_severity,
                f'{finding_code_classifier}.critical_{OID_TO_CODE_NAME[o]}_extension'
            ) for o, c in criticality_mapping.items() if not c and non_critical_adherence_word is not Rfc2119Word.MAY
        }

        validations = list(self._expected_critical_extensions.values()) + list(
            self._expected_non_critical_extensions.values())

        super().__init__(validations=validations, pdu_class=rfc5280.Extension)

    def validate(self, node):
        ext_oid = node.children['extnID'].pdu

        actual_criticality = bool(node.children['critical'].pdu)

        if actual_criticality:
            non_critical_finding = self._expected_non_critical_extensions.get(ext_oid)

            if non_critical_finding is not None:
                raise validation.ValidationFindingEncountered(non_critical_finding)
        else:
            critical_finding = self._expected_critical_extensions.get(ext_oid)

            if critical_finding is not None:
                raise validation.ValidationFindingEncountered(critical_finding)


class AuthorityInformationAccessAccessMethodPresenceValidator(ElementIdentifierAllowanceValidator):
    @staticmethod
    def _retrieve_access_method_id(node):
        return node.children['accessMethod']

    def __init__(self, access_method_allowances, finding_code_classifier: str,
                 unknown_access_method_allowance: Rfc2119Word):
        unknown_access_method_finding = None if unknown_access_method_allowance == Rfc2119Word.MAY else (
            validation.ValidationFinding(
                unknown_access_method_allowance.to_severity,
                f'{finding_code_classifier}.unknown_aia_access_method_present'
            )
        )

        super().__init__('access method',
                         AuthorityInformationAccessAccessMethodPresenceValidator._retrieve_access_method_id,
                         access_method_allowances,
                         f'{finding_code_classifier}.{{oid}}_aia_access_method_present',
                         f'{finding_code_classifier}.{{oid}}_aia_access_method_absent',
                         unknown_access_method_finding, pdu_class=rfc5280.AuthorityInfoAccessSyntax)


class AttributeIdentifierAllowanceValidator(ElementIdentifierAllowanceValidator):
    @staticmethod
    def _retrieve_attribute_type_id(node):
        # assume one ATV per RDN
        return node.navigate('0.type')

    def __init__(self, attribute_allowances, finding_code_classifier: str, unknown_attribute_allowance: Rfc2119Word):
        unexpected_attribute_finding = None if unknown_attribute_allowance == Rfc2119Word.MAY else (
            validation.ValidationFinding(
                unknown_attribute_allowance.to_severity,
                finding_code_classifier + '.unknown_attribute_present'
            )
        )

        super().__init__('attribute', AttributeIdentifierAllowanceValidator._retrieve_attribute_type_id,
                         attribute_allowances,
                         finding_code_classifier + '.{oid}_attribute_present',
                         finding_code_classifier + '.{oid}_attribute_absent',
                         unexpected_attribute_finding,
                         path='certificate.tbsCertificate.subject.rdnSequence')



#!/usr/bin/env python

import argparse
import sys

from pkilint import loader, report, util, finding_filter
from pkilint.cabf import serverauth
from pkilint.cabf.serverauth import serverauth_constants
from pkilint.pkix import certificate

_CERTIFICATE_TYPE_OPTIONS = [str(t).replace('_', '-') for t in serverauth_constants.CertificateType]


class ServerauthCertificateTypeAction(argparse.Action):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        cert_type = serverauth_constants.CertificateType.from_option_str(values)

        setattr(namespace, self.dest, cert_type)


def main(cli_args=None) -> int:
    parser = argparse.ArgumentParser(
        description=f'CA/Browser Forum TLS Baseline Requirements v{serverauth_constants.BR_VERSION} Certificate Linter'
    )

    subparsers = parser.add_subparsers(dest='command', required=True)

    validations_parser = subparsers.add_parser('validations',
                                               help='Output the set of validations which this linter performs')
    validations_parser.add_argument('-t', '--type', required=True,
                                    type=str.upper,
                                    action=ServerauthCertificateTypeAction,
                                    help='The type of certificate',
                                    choices=_CERTIFICATE_TYPE_OPTIONS)

    lint_parser = subparsers.add_parser('lint', help='Lint the specified certificate')

    detect_options_group = lint_parser.add_mutually_exclusive_group(required=True)
    detect_options_group.add_argument('-d', '--detect', action='store_true',
                                      help='Detect the type of certificate from reserved CA/B Forum policy '
                                           'OID, EKU(s), name constraints, and basic constraints.')
    detect_options_group.add_argument('-t', '--type',
                                      type=str.upper,
                                      action=ServerauthCertificateTypeAction,
                                      help='The type of certificate',
                                      choices=_CERTIFICATE_TYPE_OPTIONS)
    lint_parser.add_argument('-o', '--output', action='store_true',
                             help='Output the type of certificate to standard error. This option may be '
                                  'useful when using the --detect option.')
    lint_parser.add_argument('-r', '--report-all', action='store_true', help='Report all findings without filtering '
                             'any PKIX findings that are superseded by CA/Browser Forum requirements')

    util.add_standard_args(lint_parser)
    lint_parser.add_argument('file', type=argparse.FileType('rb'),
                             help='The certificate to lint'
                             )

    args = parser.parse_args(cli_args)

    if args.command == 'validations':
        doc_validator = certificate.create_pkix_certificate_validator_container(
            serverauth.create_decoding_validators(),
            serverauth.create_validators(args.type)
        )

        print(report.report_included_validations(doc_validator))

        return 0
    else:
        cert = loader.load_certificate(args.file, args.file.name)

        if args.type:
            certificate_type = args.type
        else:
            certificate_type = serverauth.determine_certificate_type(cert)

        if args.output:
            print(certificate_type.to_option_str, file=sys.stderr)

        doc_validator = certificate.create_pkix_certificate_validator_container(
            serverauth.create_decoding_validators(),
            serverauth.create_validators(certificate_type)
        )

        results = doc_validator.validate(cert.root)

        if not args.report_all:
            results, _ = finding_filter.filter_results(
                serverauth.create_serverauth_finding_filters(certificate_type), results
            )

        print(args.format(results, args.severity))

        return report.get_findings_count(results, args.severity)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

import argparse
import re
import sys

from pyasn1.type.univ import ObjectIdentifier

from pkilint import util, loader, report
from pkilint.cabf import smime
from pkilint.cabf.smime import smime_constants
from pkilint.pkix import certificate

_CERTIFICATE_TYPES = {}
for g in smime_constants.Generation:
    for v in smime_constants.ValidationLevel:
        _CERTIFICATE_TYPES[f'{v}-{g}'] = (v, g,)

_CONFIG_FILE_REGEX = re.compile(r'(?P<oid>\d+(\.\d+)*)\D.*(?P<type>(' + '|'.join(_CERTIFICATE_TYPES.keys()) + '))')


class SmimeCertificateTypeAction(argparse.Action):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        validation_level_str, generation_str = values.split('-', maxsplit=1)

        validation_level = smime_constants.ValidationLevel[validation_level_str]
        generation = smime_constants.Generation[generation_str]

        setattr(namespace, self.dest, (validation_level, generation,))


class MappingFileAction(argparse.Action):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        mappings = {}

        for line_num, line in enumerate(values.readlines()):
            line = line.strip().upper()

            m = _CONFIG_FILE_REGEX.search(line)

            if m is None:
                raise argparse.ArgumentError(self, f'Syntax error on mapping file line {line_num + 1}: "{line}"')

            oid = ObjectIdentifier(m.group('oid'))
            validation_level, generation = _CERTIFICATE_TYPES[m.group('type')]

            mappings[oid] = (validation_level, generation,)

        setattr(namespace, self.dest, mappings)


def main(cli_args=None) -> int:
    parser = argparse.ArgumentParser(
        description=f'CA/Browser Forum S/MIME Baseline Requirements v{smime_constants.BR_VERSION} Certificate Linter'
    )

    subparsers = parser.add_subparsers(dest='command', required=True)

    validations_parser = subparsers.add_parser('validations',
                                               help='Output the set of validations which this linter performs')
    validations_parser.add_argument('-t', '--type', required=True, type=str.upper, action=SmimeCertificateTypeAction,
                                    help='The type (validation level and generation) of S/MIME certificate',
                                    choices=list(_CERTIFICATE_TYPES.keys()))

    lint_parser = subparsers.add_parser('lint', help='Lint the specified certificate')

    detect_options_group = lint_parser.add_mutually_exclusive_group(required=True)

    detect_options_group.add_argument('-d', '--detect', action='store_true',
                                      help='Detect the type of S/MIME certificate from reserved CA/B Forum policy '
                                           'OID. If the type cannot be detected, then refuse to lint the certificate.')
    detect_options_group.add_argument('-g', '--guess', action='store_true',
                                      help='Detect the type of S/MIME certificate from reserved CA/B Forum policy '
                                           'OID. If the type cannot be detected, then use heuristics to determine '
                                           'the type of S/MIME certificate.')
    detect_options_group.add_argument('-t', '--type', type=str.upper, action=SmimeCertificateTypeAction,
                                      help='The type (validation level and generation) of S/MIME certificate',
                                      choices=list(_CERTIFICATE_TYPES.keys()))

    lint_parser.add_argument('-m', '--mapping', type=argparse.FileType('r'), action=MappingFileAction,
                             help='Mapping file which contains OID to validation level and '
                                  'generation mappings. Each line of the mapping file starts with a policy OID '
                                  'followed by a non-numeric character and the certificate type to which the OID '
                                  'maps (see -t/--type option for possible values)', default=None
                             )
    lint_parser.add_argument('-o', '--output', action='store_true',
                             help='Output the type of S/MIME certificate to standard error. This option may be '
                                  'useful when using the --detect, --guess, or --mapping options.')

    util.add_standard_args(lint_parser)

    lint_parser.add_argument('file',
                             type=argparse.FileType('rb'),
                             help='The certificate to lint'
                             )

    args = parser.parse_args(cli_args)

    if args.command == 'validations':
        validation_level, generation = args.type

        doc_validator = certificate.create_pkix_certificate_validator_container(
            smime.create_decoding_validators(),
            smime.create_subscriber_validators(
                validation_level, generation
            )
        )

        print(report.report_included_validations(doc_validator))

        return 0
    else:
        cert = loader.load_certificate(args.file, args.file.name)

        if args.detect:
            v_g = smime.determine_validation_level_and_generation(cert, args.mapping)

            if v_g is None:
                raise ValueError('Could not determine validation level and generation')
            else:
                validation_level, generation = v_g
        elif args.guess:
            validation_level, generation = smime.guess_validation_level_and_generation(cert, args.mapping)
        else:
            validation_level, generation = args.type

        if args.output:
            print(f'{validation_level}-{generation}', file=sys.stderr)

        doc_validator = certificate.create_pkix_certificate_validator_container(
            smime.create_decoding_validators(),
            smime.create_subscriber_validators(
                validation_level, generation
            )
        )

        results = doc_validator.validate(cert.root)

        print(args.format(results, args.severity))

        return report.get_findings_count(results, args.severity)


if __name__ == '__main__':
    sys.exit(main())



from pyasn1_alt_modules import rfc6960

from pkilint import document, validation


class OCSPResponseDecoder(document.ValueDecoder):
    def __init__(self, *, type_mappings):
        super().__init__(type_path='responseType',
                         value_path='response', type_mappings=type_mappings
                         )


class OCSPResponseDecodingValidator(validation.DecodingValidator):
    def __init__(self, *, decode_func, **kwargs):
        super().__init__(pdu_class=rfc6960.ResponseBytes,
                         decode_func=decode_func,
                         **kwargs
                         )


class OCSPResponseStatusValidator(validation.Validator):
    VALIDATION_ERROR_RESPONSE_HAS_BYTES = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.ocsp_error_response_has_response_body'
    )

    VALIDATION_STATUS_NUMBER_4 = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.ocsp_response_status_code_4'
    )

    def __init__(self):
        super().__init__(validations=[
            self.VALIDATION_ERROR_RESPONSE_HAS_BYTES,
            self.VALIDATION_STATUS_NUMBER_4,
        ],
            pdu_class=rfc6960.OCSPResponseStatus
        )

    def validate(self, node):
        findings = []

        if node.pdu == rfc6960.OCSPResponseStatus('undefinedStatus'):
            findings.append(validation.ValidationFindingDescription(
                self.VALIDATION_STATUS_NUMBER_4,
                None
            ))

        if (
                node.pdu != rfc6960.OCSPResponseStatus('successful') and
                'responseBytes' in node.parent.children
        ):
            findings.append(validation.ValidationFindingDescription(
                self.VALIDATION_ERROR_RESPONSE_HAS_BYTES,
                None
            ))

        return validation.ValidationResult(self, node, findings)


class OCSPResponseIsBasicValidator(validation.Validator):
    VALIDATION_RESPONSE_IS_NOT_BASIC = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.ocsp_response_is_not_basic'
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_RESPONSE_IS_NOT_BASIC],
            pdu_class=rfc6960.ResponseBytes
        )

    def validate(self, node):
        type_oid = node.children['responseType'].pdu

        if type_oid != rfc6960.id_pkix_ocsp_basic:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_RESPONSE_IS_NOT_BASIC,
                f'Unknown response type: {str(type_oid)}'
            )



import ipaddress
import operator
from datetime import timedelta

from iso3166 import countries_by_alpha2
from pyasn1_alt_modules import rfc5280, rfc6962, rfc5480

import pkilint.common
from pkilint import validation, document, oid, common
from pkilint.cabf import cabf_constants
from pkilint.cabf.asn1 import ev_guidelines
from pkilint.cabf.cabf_constants import REGISTRATION_SCHEMES
from pkilint.cabf.serverauth import serverauth_constants
from pkilint.itu import x520_name, bitstring
from pkilint.pkix import Rfc2119Word, general_name, time


class CABFOrganizationIdentifierExtensionValidator(validation.Validator):
    """Validates that the content of the CA/B Forum organizationIdentifier extension conforms with EVG 9.8.2."""

    VALIDATION_ORGANIZATION_ID_INVALID_SCHEME = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.organization_identifier_ext_invalid_registration_scheme'
    )

    VALIDATION_ORGANIZATION_ID_INVALID_COUNTRY = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.organization_identifier_ext_invalid_country'
    )

    VALIDATION_ORGANIZATION_ID_INVALID_SP = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.organization_identifier_ext_invalid_state_province_for_scheme'
    )

    def __init__(self):
        super().__init__(
            pdu_class=ev_guidelines.CABFOrganizationIdentifier,
            validations=[
                self.VALIDATION_ORGANIZATION_ID_INVALID_SCHEME,
                self.VALIDATION_ORGANIZATION_ID_INVALID_COUNTRY,
                self.VALIDATION_ORGANIZATION_ID_INVALID_SP,
            ]
        )

    def validate(self, node):
        scheme = str(node.children['registrationSchemeIdentifier'].pdu)
        country = str(node.children['registrationCountry'].pdu).upper()

        scheme_node = node.children.get('registrationStateOrProvince')
        sp = None if scheme_node is None else str(scheme_node.pdu)

        scheme_info = REGISTRATION_SCHEMES.get(scheme)

        if scheme_info is None:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_ORGANIZATION_ID_INVALID_SCHEME,
                f'Invalid registration scheme: {scheme}'
            )

        if scheme_info.country_identifier_type == cabf_constants.RegistrationSchemeCountryIdentifierType.NONE:
            valid_country_code = (country == '')
        elif scheme_info.country_identifier_type == cabf_constants.RegistrationSchemeCountryIdentifierType.XG:
            valid_country_code = (country == 'XG')
        elif scheme_info.country_identifier_type == cabf_constants.RegistrationSchemeCountryIdentifierType.ISO3166:
            valid_country_code = country in countries_by_alpha2
        else:
            raise ValueError(f'Unknown country identifier type for scheme "{scheme}": '
                             f'{scheme_info.country_identifier_type}')

        if not valid_country_code:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_ORGANIZATION_ID_INVALID_COUNTRY,
                f'Invalid country code for scheme "{scheme}": {country}'
            )

        if sp is not None and not scheme_info.allow_state_province:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_ORGANIZATION_ID_INVALID_SP,
                f'Scheme "{scheme}" does not allow state/province values'
            )


class EvSubscriberAttributeAllowanceValidator(pkilint.common.AttributeIdentifierAllowanceValidator):
    """Validates that the content of the subject conforms to EVG 9.2."""

    _CODE_CLASSIFIER = 'cabf.ev_guidelines'

    _ATTRIBUTE_ALLOWANCES = {
        rfc5280.id_at_countryName: Rfc2119Word.MUST,
        rfc5280.id_at_stateOrProvinceName: Rfc2119Word.MAY,
        rfc5280.id_at_localityName: Rfc2119Word.MAY,
        rfc5280.id_at_serialNumber: Rfc2119Word.MUST,
        x520_name.id_at_businessCategory: Rfc2119Word.MUST,
        rfc5280.id_at_organizationName: Rfc2119Word.MUST,
        x520_name.id_at_organizationIdentifier: Rfc2119Word.MAY,
        x520_name.id_at_postalCode: Rfc2119Word.MAY,
        x520_name.id_at_streetAddress: Rfc2119Word.MAY,
        ev_guidelines.id_evat_jurisdiction_countryName: Rfc2119Word.MUST,
        ev_guidelines.id_evat_jurisdiction_stateOrProvinceName: Rfc2119Word.MAY,
        ev_guidelines.id_evat_jurisdiction_localityName: Rfc2119Word.MAY,
        rfc5280.id_at_commonName: Rfc2119Word.SHOULD_NOT,
    }

    def __init__(self):
        super().__init__(self._ATTRIBUTE_ALLOWANCES, self._CODE_CLASSIFIER, Rfc2119Word.MUST_NOT)


class DvSubcriberAttributeAllowanceValidator(pkilint.common.AttributeIdentifierAllowanceValidator):
    """Validates that the content of the subject field conforms with BR 7.1.7.2."""

    _CODE_CLASSIFIER = 'cabf.serverauth.dv'

    _ATTRIBUTE_ALLOWANCES = {
        rfc5280.id_at_countryName: Rfc2119Word.MAY,
        rfc5280.id_at_commonName: Rfc2119Word.SHOULD_NOT,
    }

    def __init__(self):
        super().__init__(self._ATTRIBUTE_ALLOWANCES, self._CODE_CLASSIFIER, Rfc2119Word.MUST_NOT)


class IvSubscriberAttributeAllowanceValidator(pkilint.common.AttributeIdentifierAllowanceValidator):
    """Validates that the content of the subject field conforms with BR 7.1.7.3."""

    _CODE_CLASSIFIER = 'cabf.serverauth.iv'

    _ATTRIBUTE_ALLOWANCES = {
        rfc5280.id_at_countryName: Rfc2119Word.MUST,
        rfc5280.id_at_stateOrProvinceName: Rfc2119Word.MAY,
        rfc5280.id_at_localityName: Rfc2119Word.MAY,
        x520_name.id_at_postalCode: Rfc2119Word.SHOULD_NOT,
        x520_name.id_at_streetAddress: Rfc2119Word.SHOULD_NOT,
        rfc5280.id_at_organizationName: Rfc2119Word.SHOULD_NOT,
        rfc5280.id_at_surname: Rfc2119Word.MUST,
        rfc5280.id_at_givenName: Rfc2119Word.MUST,
        rfc5280.id_at_organizationalUnitName: Rfc2119Word.MUST_NOT,
        rfc5280.id_at_commonName: Rfc2119Word.SHOULD_NOT,
    }

    def __init__(self):
        super().__init__(self._ATTRIBUTE_ALLOWANCES, self._CODE_CLASSIFIER, Rfc2119Word.SHOULD_NOT)


class OvSubscriberAttributeAllowanceValidator(pkilint.common.AttributeIdentifierAllowanceValidator):
    """Validates that the content of the subject field conforms with BR 7.1.7.4."""

    _CODE_CLASSIFIER = 'cabf.serverauth.ov'

    _ATTRIBUTE_ALLOWANCES = {
        rfc5280.id_domainComponent: Rfc2119Word.MAY,
        rfc5280.id_at_countryName: Rfc2119Word.MUST,
        rfc5280.id_at_stateOrProvinceName: Rfc2119Word.MAY,
        rfc5280.id_at_localityName: Rfc2119Word.MAY,
        x520_name.id_at_postalCode: Rfc2119Word.SHOULD_NOT,
        x520_name.id_at_streetAddress: Rfc2119Word.SHOULD_NOT,
        rfc5280.id_at_organizationName: Rfc2119Word.MUST,
        rfc5280.id_at_surname: Rfc2119Word.MUST_NOT,
        rfc5280.id_at_givenName: Rfc2119Word.MUST_NOT,
        rfc5280.id_at_organizationalUnitName: Rfc2119Word.MUST_NOT,
        rfc5280.id_at_commonName: Rfc2119Word.SHOULD_NOT,
    }

    def __init__(self):
        super().__init__(self._ATTRIBUTE_ALLOWANCES, self._CODE_CLASSIFIER, Rfc2119Word.SHOULD_NOT)


# EVG 9.2.6, BR 7.1.2.7.3, BR 7.1.2.7.4
class IdentityCertificateStateProvinceAndLocalityPresenceValidator(validation.Validator):
    """Validates that the stateOrProvinceName and/or localityName subject attributes are present, as per
    EVG 9.2.6, BR 7.1.2.7.3, and BR 7.1.2.7.4."""

    VALIDATION_STATEPROVINCE_AND_LOCALITY_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.subscriber_stateprovince_and_locality_missing'
    )

    def __init__(self):
        super().__init__(validations=self.VALIDATION_STATEPROVINCE_AND_LOCALITY_MISSING,
                         path='certificate.tbsCertificate.subject'
                         )

    def validate(self, node):
        stp_attr = node.document.get_subject_attributes_by_type(rfc5280.id_at_stateOrProvinceName)
        loc_attr = node.document.get_subject_attributes_by_type(rfc5280.id_at_localityName)

        if not any(stp_attr) and not any(loc_attr):
            raise validation.ValidationFindingEncountered(self.VALIDATION_STATEPROVINCE_AND_LOCALITY_MISSING)


class EvSubscriberJurisdictionPresenceValidator(validation.Validator):
    """Validates that jurisdictionStateOrProvinceName is present when jurisdictionLocalityName is present,
    as per EVG 9.2.4."""

    VALIDATION_JURIS_STP_ABSENT_LOCALITY_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.ev_guidelines.jurisdiction_locality_present_stateprovince_missing'
    )

    def __init__(self):
        super().__init__(validations=self.VALIDATION_JURIS_STP_ABSENT_LOCALITY_PRESENT,
                         path='certificate.tbsCertificate.subject')

    def validate(self, node):
        juris_stp_attr = node.document.get_subject_attributes_by_type(
            ev_guidelines.id_evat_jurisdiction_stateOrProvinceName)
        juris_loc_attr = node.document.get_subject_attributes_by_type(ev_guidelines.id_evat_jurisdiction_localityName)

        if any(juris_loc_attr) and not any(juris_stp_attr):
            raise validation.ValidationFindingEncountered(self.VALIDATION_JURIS_STP_ABSENT_LOCALITY_PRESENT)


class SubscriberExtensionAllowanceValidator(pkilint.common.ExtensionIdentifierAllowanceValidator):
    """Validates that the included extensions conform with BR 7.1.2.7.6."""

    _CODE_CLASSIFIER = 'cabf.serverauth.subscriber'

    _EXTENSION_ALLOWANCES = {
        rfc5280.id_pe_authorityInfoAccess: Rfc2119Word.MUST,
        rfc5280.id_ce_authorityKeyIdentifier: Rfc2119Word.MUST,
        rfc5280.id_ce_certificatePolicies: Rfc2119Word.MUST,
        rfc5280.id_ce_extKeyUsage: Rfc2119Word.MUST,
        rfc5280.id_ce_subjectAltName: Rfc2119Word.MUST,
        rfc5280.id_ce_nameConstraints: Rfc2119Word.MUST_NOT,
        rfc5280.id_ce_keyUsage: Rfc2119Word.SHOULD,
        rfc5280.id_ce_basicConstraints: Rfc2119Word.MAY,
        rfc5280.id_ce_cRLDistributionPoints: Rfc2119Word.MAY,
        rfc5280.id_ce_subjectKeyIdentifier: Rfc2119Word.SHOULD_NOT,
    }

    def __init__(self, certificate_type):
        allowances = self._EXTENSION_ALLOWANCES.copy()

        if certificate_type in serverauth_constants.SUBSCRIBER_PRECERT_TYPES:
            allowances.update({
                rfc6962.id_ce_criticalPoison: Rfc2119Word.MUST,
                rfc6962.id_ce_embeddedSCT: Rfc2119Word.MUST_NOT,
            })
        elif certificate_type in serverauth_constants.SUBSCRIBER_FINAL_CERTIFICATE_TYPES:
            allowances.update({
                rfc6962.id_ce_criticalPoison: Rfc2119Word.MUST_NOT,
                rfc6962.id_ce_embeddedSCT: Rfc2119Word.MAY,
            })
        else:
            raise ValueError(f'Unsupported certificate type: {certificate_type}')

        super().__init__(allowances, self._CODE_CLASSIFIER, Rfc2119Word.SHOULD_NOT)


class SubscriberEkuAllowanceValidator(pkilint.common.ExtendedKeyUsageAllowanceValidator):
    """Validates that the content of the extended key usage extension conforms with BR 7.1.2.7.10."""

    _CODE_CLASSIFIER = 'cabf.serverauth.subscriber'

    _EKU_ALLOWANCES = {
        rfc5280.id_kp_serverAuth: Rfc2119Word.MUST,
        rfc5280.id_kp_clientAuth: Rfc2119Word.MAY,
        rfc5280.id_kp_codeSigning: Rfc2119Word.MUST_NOT,
        rfc5280.id_kp_emailProtection: Rfc2119Word.MUST_NOT,
        rfc5280.id_kp_timeStamping: Rfc2119Word.MUST_NOT,
        rfc5280.id_kp_OCSPSigning: Rfc2119Word.MUST_NOT,
        rfc5280.anyExtendedKeyUsage: Rfc2119Word.MUST_NOT,
        rfc6962.id_kp_precertificateSigning: Rfc2119Word.MUST_NOT,
    }

    def __init__(self):
        super().__init__(self._EKU_ALLOWANCES, self._CODE_CLASSIFIER, Rfc2119Word.SHOULD_NOT)


class SubscriberKeyUsageValidator(validation.Validator):
    """Validates that the content of the key usage extension conforms with BR 7.1.2.7.11."""

    VALIDATION_REQUIRED_KU_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.subscriber_required_ku_missing'
    )

    VALIDATION_RECOMMENDED_KU_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'cabf.serverauth.subscriber_recommended_ku_missing'
    )

    VALIDATION_PROHIBITED_KU_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.subscriber_prohibited_ku_present'
    )

    VALIDATION_DISCOURAGED_KU_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'cabf.serverauth.subscriber_discouraged_ku_present'
    )

    VALIDATION_RSA_DIGSIG_AND_KEYENCIPHERMENT_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'cabf.serverauth.subscriber_rsa_digitalsignature_and_keyencipherment_present'
    )

    _SPKI_OID_TO_KU_ALLOWANCES_MAPPING = {
        rfc5480.rsaEncryption: {
            'digitalSignature': Rfc2119Word.SHOULD,
            'keyEncipherment': Rfc2119Word.MAY,
            'dataEncipherment': Rfc2119Word.SHOULD_NOT,
        },
        rfc5480.id_ecPublicKey: {
            'digitalSignature': Rfc2119Word.MUST,
            'keyAgreement': Rfc2119Word.SHOULD_NOT,
        },
    }

    def __init__(self):
        super().__init__(validations=[self.VALIDATION_REQUIRED_KU_MISSING, self.VALIDATION_PROHIBITED_KU_PRESENT,
                                      self.VALIDATION_RECOMMENDED_KU_MISSING, self.VALIDATION_DISCOURAGED_KU_PRESENT,
                                      self.VALIDATION_RSA_DIGSIG_AND_KEYENCIPHERMENT_PRESENT],
                         pdu_class=rfc5280.KeyUsage)

    def validate(self, node):
        spki_alg_oid = node.navigate(':certificate.tbsCertificate.subjectPublicKeyInfo.algorithm.algorithm').pdu

        allowances = self._SPKI_OID_TO_KU_ALLOWANCES_MAPPING.get(spki_alg_oid)

        if allowances is None:
            # unsupported SPKI, just return and let another validator throw the error
            return

        warning_findings = []

        for k in rfc5280.KeyUsage.namedValues:
            k = str(k)

            requirement_word = allowances.get(k)

            if bitstring.has_named_bit(node, k):
                if requirement_word is None:
                    raise validation.ValidationFindingEncountered(
                        self.VALIDATION_PROHIBITED_KU_PRESENT,
                        f'Prohibited KU present: {k}'
                    )
                elif requirement_word == Rfc2119Word.SHOULD_NOT:
                    warning_findings.append(validation.ValidationFindingDescription(
                        self.VALIDATION_DISCOURAGED_KU_PRESENT,
                        f'Discouraged KU present: {k}'
                    ))
            else:
                if requirement_word == Rfc2119Word.MUST:
                    raise validation.ValidationFindingEncountered(
                        self.VALIDATION_REQUIRED_KU_MISSING,
                        f'Required KU missing: {k}'
                    )
                elif requirement_word == Rfc2119Word.SHOULD:
                    warning_findings.append(validation.ValidationFindingDescription(
                        self.VALIDATION_RECOMMENDED_KU_MISSING,
                        f'Recommended KU missing: {k}'
                    ))

        if spki_alg_oid == rfc5480.rsaEncryption and (
                bitstring.has_named_bit(node, 'digitalSignature') and
                bitstring.has_named_bit(node, 'keyEncipherment')):
            warning_findings.append(
                validation.ValidationFindingDescription(self.VALIDATION_RSA_DIGSIG_AND_KEYENCIPHERMENT_PRESENT, None)
            )

        return validation.ValidationResult(self, node, warning_findings)


class SubscriberSanGeneralNameTypeValidator(validation.Validator):
    """Validates that the types of GeneralNames included in the SAN extension conform to BR 7.1.2.7.12."""

    VALIDATION_PROHIBITED_SAN_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.prohibited_san_type'
    )

    def __init__(self):
        super().__init__(validations=self.VALIDATION_PROHIBITED_SAN_TYPE, pdu_class=rfc5280.GeneralName,
                         predicate=lambda n: n.parent is not None and isinstance(n.parent.pdu, rfc5280.SubjectAltName))

    def validate(self, node):
        gn_type, _ = node.child

        if gn_type not in {'dNSName', 'iPAddress'}:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PROHIBITED_SAN_TYPE,
                f'Prohibited SAN GeneralName type: {gn_type}'
            )


class EvSanGeneralNameTypeValidator(validation.Validator):
    """Validates that the types of GeneralNames included in the SAN extension conform to EVG 9.8.1."""

    VALIDATION_PROHIBITED_SAN_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.ev_guidelines.prohibited_san_type'
    )

    def __init__(self):
        super().__init__(validations=self.VALIDATION_PROHIBITED_SAN_TYPE, pdu_class=rfc5280.GeneralName,
                         predicate=lambda n: n.parent is not None and isinstance(n.parent.pdu, rfc5280.SubjectAltName))

    def validate(self, node):
        gn_type, _ = node.child

        if gn_type != 'dNSName':
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PROHIBITED_SAN_TYPE,
                f'Prohibited SAN GeneralName type: {gn_type}'
            )


class SubscriberValidityPeriodValidator(time.ValidityPeriodThresholdsValidator):
    """Validates that the validity period conforms to BR 7.1.2.7."""

    _THRESHOLDS = [
        (
            operator.le,
            timedelta(days=398),
            validation.ValidationFinding(
                validation.ValidationFindingSeverity.ERROR,
                'cabf.certificate_validity_period_exceeds_398_days'
            )
        ),
        (
            operator.le,
            timedelta(days=397),
            validation.ValidationFinding(
                validation.ValidationFindingSeverity.WARNING,
                'cabf.certificate_validity_period_exceeds_397_days'
            )
        )
    ]

    def __init__(self):
        super().__init__(end_validity_node_retriever=lambda n: n.navigate('^.notAfter'),
                         inclusive_second=True,
                         validity_period_thresholds=self._THRESHOLDS,
                         path='certificate.tbsCertificate.validity.notBefore'
                         )


class SubscriberCommonNameValidator(validation.Validator):
    """Validates that the content of the commonName attribute conforms to BR 7.1.4.3."""

    VALIDATION_COMMON_NAME_UNKNOWN_SOURCE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.subscriber_common_name_unknown_source'
    )

    VALIDATION_UNPARSED_CN_ENCOUNTERED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        'cabf.serverauth.unparsed_common_name_encountered'
    )

    VALIDATION_UNPARSED_SAN_EXTENSION_ENCOUNTERED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        'cabf.serverauth.unparsed_san_extension_encountered'
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_COMMON_NAME_UNKNOWN_SOURCE, self.VALIDATION_UNPARSED_CN_ENCOUNTERED,
                         self.VALIDATION_UNPARSED_SAN_EXTENSION_ENCOUNTERED],
            pdu_class=rfc5280.X520CommonName)

    def validate(self, node):
        # unparsed CN, return
        if not any(node.children):
            raise validation.ValidationFindingEncountered(self.VALIDATION_UNPARSED_CN_ENCOUNTERED)

        _, value_node = node.child
        value_str = str(value_node.pdu)

        san_ext_and_idx = node.document.get_extension_by_oid(rfc5280.id_ce_subjectAltName)

        if san_ext_and_idx is None:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_COMMON_NAME_UNKNOWN_SOURCE,
                f'Unknown source for value of common name: "{value_str}"'
            )

        san_ext_node, _ = san_ext_and_idx

        try:
            san_value_node = san_ext_node.navigate('extnValue.subjectAltName')
        except document.PDUNavigationFailedError:
            # unparsed SAN extension, return
            raise validation.ValidationFindingEncountered(self.VALIDATION_UNPARSED_SAN_EXTENSION_ENCOUNTERED)

        for gn in san_value_node.children.values():
            gn_type, gn_value = gn.child

            if gn_type == 'dNSName' and str(gn_value.pdu) == value_str:
                return
            elif gn_type == 'iPAddress':
                address_octets = gn_value.pdu.asOctets()

                if len(address_octets) == 4:
                    ip_addr = ipaddress.IPv4Address(address_octets)
                elif len(address_octets) == 16:
                    ip_addr = ipaddress.IPv6Address(address_octets)
                else:
                    # whoa Nellie! let the PKIX validator complain about this one
                    continue

                if str(ip_addr) == value_str:
                    return

        raise validation.ValidationFindingEncountered(
            self.VALIDATION_COMMON_NAME_UNKNOWN_SOURCE,
            f'Unknown source for value of common name: "{value_str}"'
        )


class SubscriberExtensionCriticalityValidator(pkilint.common.ExtensionCriticalityValidator):
    """Validates that the criticality of extensions conforms to BR 7.1.2.7.6."""

    _CODE_CLASSIFIER = 'cabf.serverauth.subscriber'

    _CRITICALITY_MAPPING = {
        rfc5280.id_pe_authorityInfoAccess: False,
        rfc5280.id_ce_authorityKeyIdentifier: False,
        rfc5280.id_ce_certificatePolicies: False,
        rfc5280.id_ce_extKeyUsage: False,
        rfc5280.id_ce_keyUsage: True,
        rfc5280.id_ce_basicConstraints: True,
        rfc5280.id_ce_cRLDistributionPoints: False,
        rfc6962.id_ce_embeddedSCT: False,
        rfc5280.id_ce_subjectKeyIdentifier: False,
    }

    def __init__(self):
        super().__init__(self._CRITICALITY_MAPPING, self._CODE_CLASSIFIER, Rfc2119Word.MUST, Rfc2119Word.MUST)


class SubscriberBasicConstraintsValidator(validation.Validator):
    VALIDATION_CA_BIT_SET = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.subscriber_basic_constraints_ca_bit_set'
    )

    def __init__(self):
        super().__init__(validations=self.VALIDATION_CA_BIT_SET, pdu_class=rfc5280.BasicConstraints)

    def validate(self, node):
        if bool(node.children['cA'].pdu):
            raise validation.ValidationFindingEncountered(self.VALIDATION_CA_BIT_SET)


class SubscriberPoliciesValidator(validation.Validator):
    """Validates that the certificate policy OID(s) conform to BR 7.1.2.7.9."""
    VALIDATION_ANYPOLICY_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.subscriber_anypolicy_oid_present'
    )

    VALIDATION_MULTIPLE_RESERVED_OIDS = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.ca_multiple_reserved_policy_oids'
    )

    VALIDATION_NO_RESERVED_OID = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.subscriber_missing_reserved_policy_oid'
    )

    VALIDATION_FIRST_OID_NOT_RESERVED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'cabf.serverauth.subscriber_first_policy_oid_not_reserved'
    )

    def __init__(self, certificate_type: serverauth_constants.CertificateType):
        if certificate_type not in serverauth_constants.SUBSCRIBER_CERTIFICATE_TYPES:
            raise ValueError(f'Unsupported certificate type: {certificate_type}')

        self._certificate_type = certificate_type

        if certificate_type in serverauth_constants.EV_CERTIFICATE_TYPES:
            self._expected_reserved_oid = serverauth_constants.ID_POLICY_EV
        elif certificate_type in serverauth_constants.OV_CERTIFICATE_TYPES:
            self._expected_reserved_oid = serverauth_constants.ID_POLICY_OV
        elif certificate_type in serverauth_constants.IV_CERTIFICATE_TYPES:
            self._expected_reserved_oid = serverauth_constants.ID_POLICY_IV
        elif certificate_type in serverauth_constants.DV_CERTIFICATE_TYPES:
            self._expected_reserved_oid = serverauth_constants.ID_POLICY_DV
        else:
            raise ValueError(f'Unsupported certificate type: {certificate_type}')

        super().__init__(validations=[self.VALIDATION_ANYPOLICY_PRESENT,
                                      self.VALIDATION_MULTIPLE_RESERVED_OIDS,
                                      self.VALIDATION_NO_RESERVED_OID,
                                      self.VALIDATION_FIRST_OID_NOT_RESERVED],
                         pdu_class=rfc5280.CertificatePolicies)

    def validate(self, node):
        policy_oids = [pi.children['policyIdentifier'].pdu for pi in node.children.values()]

        if rfc5280.anyPolicy in policy_oids:
            raise validation.ValidationFindingEncountered(self.VALIDATION_ANYPOLICY_PRESENT)

        if self._expected_reserved_oid not in policy_oids:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_NO_RESERVED_OID,
                f'Required policy OID "{str(self._expected_reserved_oid)}" missing')

        reserved_oids = set(policy_oids) & serverauth_constants.SERVERAUTH_RESERVED_POLICY_OIDS

        if len(reserved_oids) > 1:
            oids_str = oid.format_oids(reserved_oids)

            raise validation.ValidationFindingEncountered(self.VALIDATION_MULTIPLE_RESERVED_OIDS,
                                                          f'Multiple reserved policy OIDs present: {oids_str}')

        if policy_oids[0] not in serverauth_constants.SERVERAUTH_RESERVED_POLICY_OIDS:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_FIRST_OID_NOT_RESERVED
            )


class EvWildcardAllowanceValidator(validation.Validator):
    """Validates that wildcard dNSNames conform to EVG 9.8.1."""

    VALIDATION_EV_WILDCARD_SAN_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.ev_guidelines.ev_wildcard_san_present'
    )

    def __init__(self):
        super().__init__(validations=self.VALIDATION_EV_WILDCARD_SAN_PRESENT,
                         predicate=general_name.create_generalname_type_predicate('dNSName'))

    def validate(self, node):
        value = str(node.pdu)

        if value.startswith('*') and not value.lower().endswith('.onion'):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_EV_WILDCARD_SAN_PRESENT,
                f'Wildcard SAN present: "{value}"'
            )


class SubscriberAuthorityInformationAccessAccessMethodPresenceValidator(
        common.AuthorityInformationAccessAccessMethodPresenceValidator):
    """Validates that AIA access methods conform to BR 7.1.2.10.3."""

    _CODE_CLASSIFIER = 'cabf.serverauth.subscriber'

    _ACCESS_METHOD_ALLOWANCES = {
        rfc5280.id_ad_ocsp: Rfc2119Word.MUST,
        rfc5280.id_ad_caIssuers: Rfc2119Word.SHOULD,
    }

    def __init__(self):
        super().__init__(self._ACCESS_METHOD_ALLOWANCES, self._CODE_CLASSIFIER, Rfc2119Word.MUST_NOT)



from pkilint import validation
from pkilint.pkix import certificate
from pkilint.cabf import serverauth, smime


def _test(validator, context: str):
    declared_validations = {getattr(validator, a) for a in dir(validator) if a.startswith('VALIDATION_')}

    reported_validations = set(validator.validations)

    missing_validations = declared_validations - reported_validations

    assert not any(missing_validations), f'{context}: {validator.__class__.__name__} {missing_validations}'

    if isinstance(validator, validation.ValidatorContainer):
        for v in validator.validators:
            _test(v, context)


def test_serverauth():
    for cert_type in serverauth.serverauth_constants.CertificateType:
        validator = certificate.create_pkix_certificate_validator_container(
            serverauth.create_decoding_validators(),
            serverauth.create_validators(cert_type)
        )

        context = f'cabf.serverauth.{cert_type}'

        _test(validator, context)


def test_smime():
    for validation_level in smime.smime_constants.ValidationLevel:
        for generation in smime.smime_constants.Generation:
            validator = certificate.create_pkix_certificate_validator_container(
                smime.create_decoding_validators(),
                smime.create_subscriber_validators(validation_level, generation)
            )

            context = f'cabf.smime.{validation_level}-{generation}'

            _test(validator, context)



import operator
import re
from datetime import datetime, timedelta
from datetime import timezone

from dateutil.relativedelta import relativedelta
from pyasn1.type import useful
from pyasn1_alt_modules import rfc5280

from pkilint import validation
from pkilint.document import PDUNavigationFailedError

_REGEX_UTC_TIME = re.compile(
    r'^(?P<year>\d{2})(?P<month>\d{2})(?P<day>\d{2})'
    r'(?P<hour>\d{2})(?P<minute>\d{2})(?P<second>\d{2})Z$'
)

_REGEX_GENERALIZED_TIME = re.compile(
    r'^(?P<year>\d{4})(?P<month>\d{2})(?P<day>\d{2})'
    r'(?P<hour>\d{2})(?P<minute>\d{2})(?P<second>\d{2})Z$'
)

VALIDATION_INVALID_TIME_SYNTAX = validation.ValidationFinding(
    validation.ValidationFindingSeverity.ERROR,
    'pkix.invalid_time_syntax'
)


def _match_to_datetime(m):
    # check if UTCTime
    has_2_digit_year = len(m['year']) == 2

    date_components = {k: int(v) for k, v in m.groupdict().items()}

    if has_2_digit_year:
        if date_components['year'] >= 50:
            date_components['year'] += 1900
        else:
            date_components['year'] += 2000

    return datetime(date_components['year'], date_components['month'],
                    date_components['day'], date_components['hour'],
                    date_components['minute'], date_components['second'], 0,  # msec
                    timezone.utc)


def parse_generalizedtime(value):
    strval = str(value)

    m = _REGEX_GENERALIZED_TIME.match(strval)
    if m is None:
        raise ValueError(
            f'"{strval}" does not match GeneralizedTime regular '
            f'expression "{_REGEX_GENERALIZED_TIME.pattern}"'
        )

    return _match_to_datetime(m)


def parse_utctime(value):
    strval = str(value)

    m = _REGEX_UTC_TIME.match(strval)
    if m is None:
        raise ValueError(
            f'"{strval}" does not match UTCTime regular '
            f'expression "{_REGEX_UTC_TIME.pattern}"'
        )

    return _match_to_datetime(m)


def parse_time_node(value):
    if 'generalTime' in value.children:
        return parse_generalizedtime(value.children['generalTime'].pdu)
    else:
        return parse_utctime(value.children['utcTime'].pdu)


class TimeCorrectEncodingValidator(validation.Validator):
    VALIDATION_WRONG_TIME_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.wrong_time_useful_type'
    )

    def __init__(self):
        super().__init__(pdu_class=rfc5280.Time, validations=[
            VALIDATION_INVALID_TIME_SYNTAX,
            self.VALIDATION_WRONG_TIME_TYPE,
        ])

    def validate(self, node):
        try:
            parsed_datetime = parse_time_node(node)
        except ValueError as e:
            raise validation.ValidationFindingEncountered(
                VALIDATION_INVALID_TIME_SYNTAX,
                str(e)
            )

        if ('generalTime' in node.children) and (
                parsed_datetime.year >= 1950) and (
                parsed_datetime.year < 2050):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_WRONG_TIME_TYPE,
                'Time values that contain a year value of '
                f'"{parsed_datetime.year}" must be encoded using UTCTime'
            )


def _parse_date(node):
    try:
        if isinstance(node.pdu, rfc5280.Time):
            return parse_time_node(node)
        elif isinstance(node.pdu, useful.GeneralizedTime):
            return parse_generalizedtime(str(node.pdu))
        elif isinstance(node.pdu, useful.UTCTime):
            return parse_utctime(str(node.pdu))
        else:
            raise ValueError('Unsupported ASN.1 time type')
    except ValueError as e:
        raise validation.ValidationFindingEncountered(
            VALIDATION_INVALID_TIME_SYNTAX,
            f'{node.name}: {str(e)}'
        )


class ValidityPeriodDifferenceValidator(validation.Validator):
    VALIDATION_PERIOD_END_VALIDITY_NODE_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.validity_period_end_value_missing'
    )

    def __init__(self, *, end_validity_node_retriever, validations, **kwargs):
        self._end_validity_node_retriever = end_validity_node_retriever

        super().__init__(validations=[
                                         VALIDATION_INVALID_TIME_SYNTAX,
                                         self.VALIDATION_PERIOD_END_VALIDITY_NODE_MISSING,
                                     ] + validations, **kwargs)

    def validate(self, node):
        try:
            end_node = self._end_validity_node_retriever(node)
        except PDUNavigationFailedError:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PERIOD_END_VALIDITY_NODE_MISSING
            )

        start_datetime = _parse_date(node)
        end_datetime = _parse_date(end_node)

        self.validate_date_range(start_datetime, end_datetime)

    def validate_date_range(self, start_datetime, end_datetime):
        pass


class SaneValidityPeriodValidator(ValidityPeriodDifferenceValidator):
    def __init__(self, *, end_validity_node_retriever, validation, **kwargs):
        super().__init__(
            end_validity_node_retriever=end_validity_node_retriever,
            validations=[validation],
            **kwargs
        )

    def validate_date_range(self, start_datetime, end_datetime):
        if start_datetime > end_datetime:
            raise validation.ValidationFindingEncountered(
                self._validations[0],
                f'Start of validity period "{start_datetime}" is greater than '
                f'end of validity period "{end_datetime}"'
            )


class ValidityPeriodThresholdsValidator(ValidityPeriodDifferenceValidator):
    def __init__(self, *, end_validity_node_retriever, inclusive_second=False,
                 validity_period_thresholds, **kwargs
                 ):
        validity_period_thresholds.sort(key=lambda v: v[2].severity)

        validations = [v[2] for v in validity_period_thresholds]

        self._validity_period_thresholds = validity_period_thresholds
        self._inclusive_second = inclusive_second

        super().__init__(
            end_validity_node_retriever=end_validity_node_retriever,
            validations=validations,
            **kwargs
        )

    def validate_date_range(self, start_datetime, end_datetime):
        if self._inclusive_second:
            end_datetime += timedelta(seconds=1)
        validity = end_datetime - start_datetime

        for op, threshold_value, finding in self._validity_period_thresholds:
            if isinstance(threshold_value, relativedelta):
                is_valid = op(end_datetime, start_datetime + threshold_value)
            else:
                is_valid = op(validity, threshold_value)

            if not is_valid:
                if op in [operator.ge, operator.gt]:
                    raise validation.ValidationFindingEncountered(
                        finding,
                        f'Validity period of {validity} is below minimum '
                        f'value of {threshold_value}'
                    )
                else:
                    raise validation.ValidationFindingEncountered(
                        finding,
                        f'Validity period of {validity} exceeds maximum '
                        f'value of {threshold_value}'
                    )


class UtcTimeCorrectSyntaxValidator(validation.Validator):
    VALIDATION_INCORRECT_SYNTAX = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.utctime_incorrect_syntax'
    )

    def __init__(self):
        super().__init__(pdu_class=useful.UTCTime,
                         validations=[self.VALIDATION_INCORRECT_SYNTAX]
                         )

    def validate(self, node):
        try:
            parse_utctime(node.pdu)
        except ValueError as e:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INCORRECT_SYNTAX, str(e)
            )


class GeneralizedTimeCorrectSyntaxValidator(validation.Validator):
    VALIDATION_INCORRECT_SYNTAX = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.generalizedtime_incorrect_syntax'
    )

    def __init__(self):
        super().__init__(pdu_class=useful.GeneralizedTime,
                         validations=[self.VALIDATION_INCORRECT_SYNTAX]
                         )

    def validate(self, node):
        try:
            parse_generalizedtime(node.pdu)
        except ValueError as e:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INCORRECT_SYNTAX, str(e)
            )



import csv
import io
import json
from typing import Iterable, Optional, Any

from pkilint.validation import ValidationFindingSeverity, ValidationResult, ValidationFindingDescription


class ReportGeneratorBase:
    def __init__(self, results: Iterable[ValidationResult], severity_threshold: Optional[ValidationFindingSeverity],
                 report_context: Optional[Any] = None):
        self.results = results
        self.severity_threshold = severity_threshold
        self.report_context = report_context

    def get_finding_descriptions_for_result(self, result):
        return [f for f in result.finding_descriptions
                if self.severity_threshold is None or f.finding.severity <= self.severity_threshold
                ]

    def is_relevant_result(self, result):
        return self.severity_threshold is None or any(self.get_finding_descriptions_for_result(result))

    def handle_result(self, result) -> Optional[Any]:
        pass

    def handle_finding_description(self, result: ValidationResult, finding_description: ValidationFindingDescription,
                                   result_context: Optional[Any]):
        pass

    def generate(self):
        for result in self.results:
            result_context = self.handle_result(result)

            for finding_description in self.get_finding_descriptions_for_result(result):
                self.handle_finding_description(result, finding_description, result_context)


class ReportGeneratorPlaintext(ReportGeneratorBase):
    def __init__(self, results, severity_threshold):
        super().__init__(results, severity_threshold, io.StringIO())

    def handle_result(self, result):
        if self.is_relevant_result(result):
            self.report_context.write(f'{result.validator} @ {result.node.path}\n')

    def handle_finding_description(self, result: ValidationResult, finding_description: ValidationFindingDescription,
                                   result_context: Optional[Any]):
        self.report_context.write(f'    {finding_description}\n')

    def generate(self):
        super().generate()

        return self.report_context.getvalue()


class ReportGeneratorCsv(ReportGeneratorBase):
    _CSV_FIELDNAMES = ['node_path', 'validator', 'severity', 'code', 'message']

    def __init__(self, results, severity_threshold, output_headers=True):
        self._output_io = io.StringIO()

        output_csv = csv.DictWriter(self._output_io, fieldnames=self._CSV_FIELDNAMES)

        if output_headers:
            output_csv.writeheader()

        super().__init__(results, severity_threshold, output_csv)

    def handle_finding_description(self, result, finding_description, result_context):
        row = {
            'node_path': result.node.path,
            'validator': str(result.validator),
            'severity': finding_description.finding.severity.name,
            'code': finding_description.finding.code,
            'message': '' if finding_description.message is None else finding_description.message
        }

        self.report_context.writerow(row)

    def generate(self):
        super().generate()

        return self._output_io.getvalue()


class ReportGeneratorJson(ReportGeneratorBase):
    def __init__(self, results, severity_threshold):
        super().__init__(results, severity_threshold, [])

    def handle_result(self, result) -> Optional[Any]:
        if self.is_relevant_result(result):
            result_dict = {
                'node_path': result.node.path,
                'validator': str(result.validator),
                'finding_descriptions': []
            }

            self.report_context.append(result_dict)

            return result_dict['finding_descriptions']

    def handle_finding_description(self, result: ValidationResult, finding_description: ValidationFindingDescription,
                                   result_context: Optional[Any]):
        result_context.append(
            {
                'severity': finding_description.finding.severity.name,
                'code': finding_description.finding.code,
                'message': finding_description.message,
            }
        )

    def generate(self):
        super().generate()

        return json.dumps({'results': self.report_context})


def get_findings_count(results: Iterable[ValidationResult],
                       severity_threshold: ValidationFindingSeverity = None
                       ):
    findings = 0
    for result in results:
        for finding in result.finding_descriptions:
            if severity_threshold is None or finding.finding.severity <= severity_threshold:
                findings += 1

    return findings


def report_wrapper(report_generator_cls, *args, **kwargs):
    report_generator = report_generator_cls(*args, **kwargs)

    return report_generator.generate()


REPORT_FORMATS = {
    'TEXT': ReportGeneratorPlaintext,
    'CSV': ReportGeneratorCsv,
    'JSON': ReportGeneratorJson,
}

_VALIDATION_LIST_CSV_FIELDNAMES = ['severity', 'code']


def report_included_validations(*args):
    s = io.StringIO()

    c = csv.DictWriter(s, fieldnames=_VALIDATION_LIST_CSV_FIELDNAMES)
    c.writeheader()

    all_validations = set()
    for validator in args:
        all_validations.update(validator.validations)

    validations = sorted(all_validations, key=lambda v: f'{int(v.severity)}-{v.code}')

    for v in validations:
        c.writerow({'severity': str(v.severity), 'code': v.code})

    return s.getvalue()



# Auto-generated by asn1ate v.0.6.0 from ev_guidelines.asn1
# (last modified on 2020-12-23 17:42:25.113525)

from pyasn1.type import univ, char, namedtype, tag, constraint
from pyasn1.type.univ import ObjectIdentifier, Choice
from pyasn1_alt_modules import rfc5280


def _OID(*components):
    output = []
    for x in tuple(components):
        if isinstance(x, univ.ObjectIdentifier):
            output.extend(list(x))
        else:
            output.append(int(x))

    return univ.ObjectIdentifier(output)


class CABFOrganizationIdentifier(univ.Sequence):
    pass


CABFOrganizationIdentifier.componentType = namedtype.NamedTypes(
    namedtype.NamedType('registrationSchemeIdentifier',
                        char.PrintableString().subtype(subtypeSpec=constraint.ValueSizeConstraint(3, 3))),
    namedtype.NamedType('registrationCountry',
                        char.PrintableString().subtype(subtypeSpec=constraint.ValueSizeConstraint(2, 2))),
    namedtype.OptionalNamedType('registrationStateOrProvince', char.PrintableString().subtype(
        subtypeSpec=constraint.ValueSizeConstraint(0, 128)).subtype(
        implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0))),
    namedtype.NamedType('registrationReference', char.UTF8String())
)

id_CABFOrganizationIdentifier = _OID(2, 23, 140, 3, 1)

id_evat_jurisdiction_localityName = ObjectIdentifier(
    '1.3.6.1.4.1.311.60.2.1.1'
)


class EVGJurisdictionLocalityName(Choice):
    pass


EVGJurisdictionLocalityName.componentType = namedtype.NamedTypes(
    namedtype.NamedType('teletexString',
                        char.TeletexString().subtype(
                            subtypeSpec=constraint.ValueSizeConstraint(1, rfc5280.ub_locality_name))),
    namedtype.NamedType('printableString', char.PrintableString().subtype(
        subtypeSpec=constraint.ValueSizeConstraint(1, rfc5280.ub_locality_name))),
    namedtype.NamedType('universalString', char.UniversalString().subtype(
        subtypeSpec=constraint.ValueSizeConstraint(1, rfc5280.ub_locality_name))),
    namedtype.NamedType('utf8String',
                        char.UTF8String().subtype(
                            subtypeSpec=constraint.ValueSizeConstraint(1, rfc5280.ub_locality_name))),
    namedtype.NamedType('bmpString',
                        char.BMPString().subtype(
                            subtypeSpec=constraint.ValueSizeConstraint(1, rfc5280.ub_locality_name)))
)

id_evat_jurisdiction_stateOrProvinceName = ObjectIdentifier(
    '1.3.6.1.4.1.311.60.2.1.2'
)


class EVGJurisdictionStateOrProvinceName(Choice):
    pass


EVGJurisdictionStateOrProvinceName.componentType = namedtype.NamedTypes(
    namedtype.NamedType('teletexString',
                        char.TeletexString().subtype(
                            subtypeSpec=constraint.ValueSizeConstraint(1, rfc5280.ub_state_name))),
    namedtype.NamedType('printableString',
                        char.PrintableString().subtype(
                            subtypeSpec=constraint.ValueSizeConstraint(1, rfc5280.ub_state_name))),
    namedtype.NamedType('universalString',
                        char.UniversalString().subtype(
                            subtypeSpec=constraint.ValueSizeConstraint(1, rfc5280.ub_state_name))),
    namedtype.NamedType('utf8String',
                        char.UTF8String().subtype(
                            subtypeSpec=constraint.ValueSizeConstraint(1, rfc5280.ub_state_name))),
    namedtype.NamedType('bmpString',
                        char.BMPString().subtype(subtypeSpec=constraint.ValueSizeConstraint(1, rfc5280.ub_state_name)))
)

id_evat_jurisdiction_countryName = ObjectIdentifier(
    '1.3.6.1.4.1.311.60.2.1.3'
)


class EVGJurisdictionCountryName(rfc5280.X520countryName):
    pass


ATTRIBUTE_TYPE_MAPPINGS = {
    id_evat_jurisdiction_countryName: EVGJurisdictionCountryName(),
    id_evat_jurisdiction_stateOrProvinceName: EVGJurisdictionStateOrProvinceName(),
    id_evat_jurisdiction_localityName: EVGJurisdictionLocalityName(),
}

EXTENSION_MAPPINGS = {
    id_CABFOrganizationIdentifier: CABFOrganizationIdentifier(),
}



#!/usr/bin/env python

import argparse

from cryptography import x509
from cryptography.hazmat.primitives.serialization import Encoding


def _convert_cert(f):
    content = f.read()

    if content.startswith(b'\x30'):
        cert = x509.load_der_x509_certificate(content)
    else:
        cert = x509.load_pem_x509_certificate(content)

    return cert.public_bytes(Encoding.PEM).decode('us-ascii')


parser = argparse.ArgumentParser()
parser.add_argument('file', type=argparse.FileType('rb'))

args = parser.parse_args()

pem = _convert_cert(args.file)

print(f'pem = """{pem}"""')



import binascii

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.asymmetric import padding, rsa, ec
from pyasn1.codec.der.encoder import encode
from pyasn1.type import univ
from pyasn1_alt_modules import rfc5280, rfc3279, rfc5480, rfc8410

from pkilint import validation, util, document
from pkilint.document import PDUNode

SUBJECT_PUBLIC_KEY_ALGORITHM_IDENTIFIER_MAPPINGS = {
    rfc3279.rsaEncryption: rfc5480.RSAPublicKey(),
    rfc5480.id_ecPublicKey: rfc5480.ECPoint(),
    rfc5480.id_ecDH: rfc5480.ECPoint(),
    rfc5480.id_ecMQV: rfc5480.ECPoint(),
}

SUBJECT_KEY_PARAMETER_ALGORITHM_IDENTIFIER_MAPPINGS = {
    rfc3279.rsaEncryption: univ.Null(),
    rfc5480.id_ecPublicKey: rfc5480.ECParameters(),
    rfc5480.id_ecDH: rfc5480.ECParameters(),
    rfc5480.id_ecMQV: rfc5480.ECParameters(),
    **{o: document.ValueDecoder.VALUE_NODE_ABSENT for o in (
        rfc8410.id_Ed448,
        rfc8410.id_Ed25519,
        rfc8410.id_X448,
        rfc8410.id_X25519,
    )}
}

EC_CURVE_OID_TO_OBJECT_MAPPINGS = {
    rfc5480.secp256r1: ec.SECP256R1(),
    rfc5480.secp384r1: ec.SECP384R1(),
    rfc5480.secp521r1: ec.SECP521R1(),
}


def convert_spki_to_object(spki_node: PDUNode):
    key_type = spki_node.navigate('algorithm.algorithm').pdu

    if key_type == rfc3279.rsaEncryption:
        modulus = spki_node.navigate(
            'subjectPublicKey.rSAPublicKey.modulus'
        ).pdu
        exponent = spki_node.navigate(
            'subjectPublicKey.rSAPublicKey.exponent'
        ).pdu

        return rsa.RSAPublicNumbers(int(modulus), int(exponent)).public_key()
    elif key_type in {rfc5480.id_ecPublicKey, rfc5480.id_ecDH, rfc5480.id_ecMQV}:
        curve_oid = spki_node.navigate(
            'algorithm.parameters.eCParameters.namedCurve'
        ).pdu

        curve = EC_CURVE_OID_TO_OBJECT_MAPPINGS.get(curve_oid)
        if curve is not None:
            return ec.EllipticCurvePublicKey.from_encoded_point(
                curve, spki_node.navigate('subjectPublicKey').pdu.asOctets()
            )

    # TODO: DSA
    return None


class SubjectPublicKeyDecoder(document.ValueDecoder):
    def __init__(self, *, type_mappings):
        super().__init__(type_path='algorithm.algorithm',
                         value_path='subjectPublicKey', type_mappings=type_mappings
                         )

    def filter_value(self, node, type_node, value_node, pdu_type):
        if isinstance(pdu_type, rfc5480.ECPoint):
            # wrap the BIT STRING in an OCTET STRING
            octet_str = univ.OctetString(value_node.pdu.asOctets())

            return encode(octet_str)
        else:
            return super().filter_value(node, type_node, value_node, pdu_type)


class SubjectPublicKeyDecodingValidator(validation.DecodingValidator):
    def __init__(self, *, decode_func, **kwargs):
        super().__init__(pdu_class=rfc5280.SubjectPublicKeyInfo,
                         decode_func=decode_func,
                         **kwargs
                         )


class SubjectPublicKeyParametersDecoder(document.ValueDecoder):
    def __init__(self, *, type_mappings):
        super().__init__(type_path='algorithm.algorithm',
                         value_path='algorithm.parameters', type_mappings=type_mappings
                         )


class SubjectPublicKeyParametersDecodingValidator(validation.DecodingValidator):
    def __init__(self, *, decode_func, **kwargs):
        super().__init__(pdu_class=rfc5280.SubjectPublicKeyInfo,
                         decode_func=decode_func,
                         **kwargs
                         )


def _calculate_method2_hash(sha1_hash):
    last_8_octets = bytearray(sha1_hash[12:])
    last_8_octets[0] = 0x40 | (last_8_octets[0] & 0xF)

    return bytes(last_8_octets)


class SubjectKeyIdentifierValidator(validation.Validator):
    VALIDATION_UNKNOWN_METHOD = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        'pkix.unknown_subject_key_identifier_calculation_method'
    )

    VALIDATION_METHOD_1 = validation.ValidationFinding(
        validation.ValidationFindingSeverity.INFO,
        'pkix.subject_key_identifier_method_1_identified'
    )

    VALIDATION_METHOD_2 = validation.ValidationFinding(
        validation.ValidationFindingSeverity.INFO,
        'pkix.subject_key_identifier_method_2_identified'
    )

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_UNKNOWN_METHOD,
                self.VALIDATION_METHOD_1,
                self.VALIDATION_METHOD_2,
            ],
            pdu_class=rfc5280.SubjectKeyIdentifier
        )

    def validate(self, node):
        public_key_node = node.document.root.navigate(
            'tbsCertificate.subjectPublicKeyInfo.subjectPublicKey'
        )

        public_key_bytes = public_key_node.pdu.asOctets()
        public_key_sha1 = util.calculate_sha1_hash(public_key_bytes)

        method2_hash = _calculate_method2_hash(public_key_sha1)

        identifier_octets = bytes(node.pdu)

        if public_key_sha1 == identifier_octets:
            finding = self.VALIDATION_METHOD_1
        elif method2_hash == identifier_octets:
            finding = self.VALIDATION_METHOD_2
        else:
            finding = self.VALIDATION_UNKNOWN_METHOD

        raise validation.ValidationFindingEncountered(finding)


def _verify_signature(public_key, message, signature,
                      signature_algorithm):
    try:
        if isinstance(public_key, rsa.RSAPublicKey):
            public_key.verify(
                signature,
                message,
                padding.PKCS1v15(),
                signature_algorithm
            )
        else:
            public_key.verify(
                signature,
                message,
                ec.ECDSA(signature_algorithm)
            )

        return True
    except InvalidSignature:
        return False


class SubjectSignatureVerificationValidator(validation.Validator):
    VALIDATION_SIGNATURE_MISMATCH = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.signature_verification_failed'
    )

    def __init__(self, *, tbs_node_retriever, **kwargs):
        super().__init__(validations=[self.VALIDATION_SIGNATURE_MISMATCH],
                         **kwargs
                         )

        self._tbs_node_retriever = tbs_node_retriever

    def validate(self, node):
        issuer_cert_doc = document.get_document_by_name(node, 'issuer')

        issuer_crypto_cert = issuer_cert_doc.cryptography_object
        subject_crypto_doc = node.document.cryptography_object
        public_key = issuer_crypto_cert.public_key()

        tbs_octets = encode(self._tbs_node_retriever(node).pdu)

        if not _verify_signature(public_key, tbs_octets,
                                 node.pdu.asOctets(),
                                 subject_crypto_doc.signature_hash_algorithm):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_SIGNATURE_MISMATCH
            )


class AllowedPublicKeyAlgorithmEncodingValidator(validation.Validator):
    def __init__(self, *, validation, allowed_encodings, **kwargs):
        super().__init__(
            validations=[validation],
            **kwargs
        )

        self._allowed_encodings = allowed_encodings

    def validate(self, node):
        encoded = encode(node.pdu)

        if encoded not in self._allowed_encodings:
            encoded_str = binascii.hexlify(encoded).decode('us-ascii')

            raise validation.ValidationFindingEncountered(
                self._validations[0],
                f'Prohibited encoding: {encoded_str}'
            )



from pkilint.etsi.asn1 import en_319_412_5, ts_119_495


ETSI_QC_STATEMENTS_MAPPINGS = {
    en_319_412_5.id_etsi_qcs_QcCompliance: None,
    en_319_412_5.id_etsi_qcs_QcLimitValue: en_319_412_5.QcEuLimitValue(),
    en_319_412_5.id_etsi_qcs_QcRetentionPeriod: en_319_412_5.QcEuRetentionPeriod(),
    en_319_412_5.id_etsi_qcs_QcSSCD: None,
    en_319_412_5.id_etsi_qcs_QcPDS: en_319_412_5.QcEuPDS(),
    en_319_412_5.id_etsi_qcs_QcType: en_319_412_5.QcType(),
    en_319_412_5.id_etsi_qcs_QcCClegislation: en_319_412_5.QcCClegislation(),
    ts_119_495.id_etsi_psd2_qcStatement: ts_119_495.PSD2QcType(),
}



from pyasn1_alt_modules import rfc5280

from pkilint import validation


class CrlDpDistributionPointValidator(validation.Validator):
    """Validates that the fields included in the CRL distribution points extension conforms with BR 7.1.2.11.2."""

    VALIDATION_CRLDP_DP_PROHIBITED_FIELD = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.crldp_dp_prohibited_field_present'
    )

    def __init__(self):
        super().__init__(validations=self.VALIDATION_CRLDP_DP_PROHIBITED_FIELD, pdu_class=rfc5280.DistributionPoint)

    def validate(self, node):
        for prohibited_field in ('reasons', 'cRLIssuer'):
            if prohibited_field in node.children:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_CRLDP_DP_PROHIBITED_FIELD,
                    f'Prohibited field present: {prohibited_field}'
                )


# BR 7.1.2.11.2
class CrlDpDistributionPointNameValidator(validation.Validator):
    """Validates that the names included in the CRL distribution points extension conforms with BR 7.1.2.11.2."""
    VALIDATION_PROHIBITED_FIELD = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.crldp_dpname_prohibited_field_present'
    )

    VALIDATION_PROHIBITED_GENERALNAME = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.crldp_dpname_prohibited_generalname_type'
    )

    VALIDATION_PROHIBITED_URI_SCHEME = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.crldp_dpname_prohibited_uri_scheme'
    )

    def __init__(self):
        super().__init__(validations=[self.VALIDATION_PROHIBITED_FIELD, self.VALIDATION_PROHIBITED_GENERALNAME,
                                      self.VALIDATION_PROHIBITED_URI_SCHEME], pdu_class=rfc5280.DistributionPointName)

    def validate(self, node):
        if 'nameRelativeToCRLIssuer' in node.children:
            raise validation.ValidationFindingEncountered(self.VALIDATION_PROHIBITED_FIELD)

        gns = node.children['fullName']

        prohibited_gn_types = {gn.child[0] for gn in gns.children.values()} - {'uniformResourceIdentifier'}

        if any(prohibited_gn_types):
            prohibited_types_str = ', '.join(prohibited_gn_types)

            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PROHIBITED_GENERALNAME, f'Prohibited GeneralName types present: {prohibited_types_str}'
            )

        for uri_node in (gn.child[1] for gn in gns.children.values()):
            scheme, _ = str(uri_node.pdu).split(':', maxsplit=1)

            if scheme.lower() != 'http':
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_PROHIBITED_URI_SCHEME,
                    f'Prohibited URI scheme: "{scheme}"'
                )


class CrlDpDistributionPointCountValidator(validation.Validator):
    """Validates that the number of distribution points conforms with BR 7.1.2.11.2."""

    VALIDATION_MULTIPLE_CRLDP_DP = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'cabf.serverauth.crldp_multiple_distributionpoints_present'
    )

    def __init__(self):
        super().__init__(validations=self.VALIDATION_MULTIPLE_CRLDP_DP, pdu_class=rfc5280.CRLDistributionPoints)

    def validate(self, node):
        if len(node.children) > 1:
            raise validation.ValidationFindingEncountered(self.VALIDATION_MULTIPLE_CRLDP_DP)


class AuthorityInformationAccessHttpUriLocationValidator(validation.Validator):
    """Validates that all locations in the AIA extension are HTTP as per BR 7.1.2.10.3, 7.1.2.7.7, and 7.1.2.8.3"""

    VALIDATION_AIA_LOCATION_NOT_URI = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.aia_location_not_uri'
    )

    VALIDATION_AIA_LOCATION_URI_NOT_HTTP = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.aia_location_uri_not_http'
    )

    def __init__(self):
        super().__init__(validations=[self.VALIDATION_AIA_LOCATION_NOT_URI, self.VALIDATION_AIA_LOCATION_URI_NOT_HTTP],
                         pdu_class=rfc5280.AccessDescription)

    def validate(self, node):
        gn_type, value = node.children['accessLocation'].child

        if gn_type != 'uniformResourceIdentifier':
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_AIA_LOCATION_NOT_URI,
                f'AIA access location is not URI: {gn_type}'
            )

        value_str = str(value.pdu)

        if not value_str.lower().startswith('http://'):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_AIA_LOCATION_URI_NOT_HTTP,
                f'AIA access location is not HTTP: "{value_str}"'
            )


# BR 7.1.2.10.3, 7.1.2.7.7, 7.1.2.8.3
class AuthorityInformationAccessUniqueLocationValidator(validation.Validator):
    """Validates that all URI locations in the AIA extension are unique, as per BR 7.1.2.10.3, 7.1.2.7.7,
    and 7.1.2.8.3."""

    VALIDATION_DUPLICATE_LOCATION_URI = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.aia_duplicate_location'
    )

    def __init__(self):
        super().__init__(validations=self.VALIDATION_DUPLICATE_LOCATION_URI,
                         pdu_class=rfc5280.AuthorityInfoAccessSyntax)

    def validate(self, node):
        uri_locations = []
        for access_description in node.children.values():
            location_type, location_value = access_description.children['accessLocation'].child

            if location_type == 'uniformResourceIdentifier':
                uri_locations.append(str(location_value.pdu))

        duplicate_locations = {l for l in uri_locations if uri_locations.count(l) > 1}

        if any(duplicate_locations):
            dup_locations_str = ', '.join(sorted(list(duplicate_locations)))

            raise validation.ValidationFindingEncountered(
                self.VALIDATION_DUPLICATE_LOCATION_URI,
                f'Duplicate AIA access locations: {dup_locations_str}'
            )


class CertificatePolicyQualifierValidator(validation.Validator):
    """Validates that the inclusion of policy qualifiers is in conformance with BR 7.1.2.3.2, 7.1.2.10.5,
    and 7.1.2.7.9."""

    VALIDATION_QUALIFIER_NOT_RECOMMENDED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'cabf.serverauth.certificate_policy_qualifier_present'
    )

    VALIDATION_PROHIBITED_QUALIFIER_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.prohibited_certificate_policy_qualifier_type'
    )

    def __init__(self):
        super().__init__(validations=[self.VALIDATION_QUALIFIER_NOT_RECOMMENDED,
                                      self.VALIDATION_PROHIBITED_QUALIFIER_TYPE], pdu_class=rfc5280.PolicyQualifierInfo)

    def validate(self, node):
        findings = [validation.ValidationFindingDescription(self.VALIDATION_QUALIFIER_NOT_RECOMMENDED, None)]

        qualifier_type_oid = node.children['policyQualifierId'].pdu

        if qualifier_type_oid != rfc5280.id_qt_cps:
            findings.append(validation.ValidationFindingDescription(
                self.VALIDATION_PROHIBITED_QUALIFIER_TYPE,
                f'Prohibited qualifier type: {str(qualifier_type_oid)}'
            ))

        return validation.ValidationResult(self, node, findings)


class EvCpsUriPresenceValidator(validation.Validator):
    """Validates that EV Subscriber certificates contain the CPS URI qualifier, as per EVG 9.7."""
    VALIDATION_EV_CPS_URI_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.ev_guidelines.cps_uri_policy_qualifier_missing'
    )

    def __init__(self):
        super().__init__(validations=self.VALIDATION_EV_CPS_URI_MISSING, pdu_class=rfc5280.CertificatePolicies)

    def validate(self, node):
        qualifier_oids = set()

        for pi in node.children.values():
            qualifiers = pi.children.get('policyQualifiers')

            if qualifiers is None:
                continue

            qualifier_oids.update((q.children['policyQualifierId'].pdu for q in qualifiers.children.values()))

        if rfc5280.id_qt_cps not in qualifier_oids:
            raise validation.ValidationFindingEncountered(self.VALIDATION_EV_CPS_URI_MISSING)



import ipaddress
import re
import typing
from urllib.parse import urlparse

import publicsuffixlist
import unicodedata
from iso3166 import countries_by_alpha2
from pyasn1_alt_modules import rfc5280, rfc8398

from pkilint import validation
from pkilint.cabf import cabf_constants
from pkilint.cabf.cabf_constants import REGISTRATION_SCHEMES
from pkilint.itu import x520_name
from pkilint.pkix import general_name


class ValidCountryCodeValidatorBase(validation.TypeMatchingValidator):
    def __init__(self, type_oid, value_path, checked_validation):
        super().__init__(type_path='type', type_oid=type_oid,
                         value_path=value_path,
                         pdu_class=rfc5280.AttributeTypeAndValue,
                         validations=[checked_validation]
                         )

    def validate_with_value(self, node, value_node):
        country_code = str(value_node.pdu).upper()

        if country_code == 'XX':
            return
        elif country_code not in countries_by_alpha2:
            raise validation.ValidationFindingEncountered(
                self.validations[0],
                f'Invalid country code: "{country_code}"'
            )


class ValidCountryValidator(ValidCountryCodeValidatorBase):
    VALIDATION_INVALID_COUNTRY_CODE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.invalid_country_code'
    )

    def __init__(self):
        super().__init__(type_oid=rfc5280.id_at_countryName,
                         value_path='value.x520countryName',
                         checked_validation=self.VALIDATION_INVALID_COUNTRY_CODE
                         )


ORG_ID_REGEX = re.compile(
    r'^(?P<scheme>[A-Z]{3})(?P<country>[a-zA-Z]{2})?(\+(?P<sp>[a-zA-Z0-9]{1,3}))?'
    r'(-(?P<reference>.+))?$'
)


class OrganizationIdentifierAttributeValidator(validation.TypeMatchingValidator):
    VALIDATION_ORGANIZATION_ID_INVALID_ENCODING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.invalid_subject_organization_identifier_encoding'
    )

    VALIDATION_ORGANIZATION_ID_INVALID_FORMAT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.invalid_subject_organization_identifier_format'
    )

    VALIDATION_ORGANIZATION_ID_INVALID_SCHEME = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.invalid_subject_organization_identifier_registration_scheme'
    )

    VALIDATION_ORGANIZATION_ID_INVALID_COUNTRY = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.invalid_subject_organization_identifier_country'
    )

    VALIDATION_ORGANIZATION_ID_INVALID_SP = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.invalid_subject_organization_identifier_state_province_for_scheme'
    )

    VALIDATION_ORGANIZATION_ID_INVALID_SP_FORMAT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.invalid_subject_organization_identifier_state_province_format'
    )

    def __init__(self, relax_stateprovince_syntax=False, additional_schemes: typing.Optional[
            typing.Mapping[str, cabf_constants.RegistrationSchemeNamingConvention]] = None):
        super().__init__(type_oid=x520_name.id_at_organizationIdentifier,
                         type_path='type', value_path='value.x520OrganizationIdentifier',
                         pdu_class=rfc5280.AttributeTypeAndValue,
                         validations=[
                             self.VALIDATION_ORGANIZATION_ID_INVALID_ENCODING,
                             self.VALIDATION_ORGANIZATION_ID_INVALID_FORMAT,
                             self.VALIDATION_ORGANIZATION_ID_INVALID_SCHEME,
                             self.VALIDATION_ORGANIZATION_ID_INVALID_COUNTRY,
                             self.VALIDATION_ORGANIZATION_ID_INVALID_SP,
                             self.VALIDATION_ORGANIZATION_ID_INVALID_SP_FORMAT,
                         ]
                         )

        if additional_schemes is None:
            additional_schemes = {}
        self._allowed_schemes = {**REGISTRATION_SCHEMES, **additional_schemes}
        self._relax_stateprovince_syntax = relax_stateprovince_syntax

    def validate_with_value(self, node, choice_node):
        name, value_node = choice_node.child

        if name not in {'utf8String', 'printableString'}:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_ORGANIZATION_ID_INVALID_ENCODING,
                f'Invalid ASN.1 encoding: {name}'
            )

        m = ORG_ID_REGEX.match(str(value_node.pdu))

        if m is None:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_ORGANIZATION_ID_INVALID_FORMAT,
                f'Invalid format: {value_node.pdu}'
            )

        scheme_info = self._allowed_schemes.get(m['scheme'])

        if scheme_info is None:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_ORGANIZATION_ID_INVALID_SCHEME,
                f'Invalid registration scheme: {m["scheme"]}'
            )

        if scheme_info.require_registration_reference and m['reference'] is None:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_ORGANIZATION_ID_INVALID_FORMAT,
                f'Missing Registration Reference: {value_node.pdu}'
            )
        elif not scheme_info.require_registration_reference and m['reference']:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_ORGANIZATION_ID_INVALID_FORMAT,
                f'Prohibited Registration Reference is present: {value_node.pdu}'
            )

        country_code = '' if m['country'] is None else m['country'].upper()

        if scheme_info.country_identifier_type == cabf_constants.RegistrationSchemeCountryIdentifierType.NONE:
            valid_country_code = (country_code == '')
        elif scheme_info.country_identifier_type == cabf_constants.RegistrationSchemeCountryIdentifierType.XG:
            valid_country_code = (country_code == 'XG')
        elif scheme_info.country_identifier_type == cabf_constants.RegistrationSchemeCountryIdentifierType.ISO3166:
            valid_country_code = (country_code in countries_by_alpha2)
        else:
            raise ValueError(f'Unknown country identifier type for scheme "{m["scheme"]}": '
                             f'{scheme_info.country_identifier_type}')

        if not valid_country_code:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_ORGANIZATION_ID_INVALID_COUNTRY,
                f'Invalid country code for scheme "{m["scheme"]}": {country_code}'
            )

        if m['sp'] is not None and not scheme_info.allow_state_province:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_ORGANIZATION_ID_INVALID_SP,
                f'Scheme "{m["scheme"]}" does not allow state/province values'
            )

        if m['sp'] is not None and not self._relax_stateprovince_syntax and not (
                len(m['sp']) == 2 and m['sp'].isalpha()):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_ORGANIZATION_ID_INVALID_SP_FORMAT,
                f'State/province "{m["sp"]}" is not two letters (will be fixed in erratum ballot)'
            )


class RelativeDistinguishedNameContainsOneElementValidator(validation.Validator):
    VALIDATION_RDN_CONTAINS_MULTIPLE_ATVS = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        'cabf.rdn_contains_multiple_atvs'
    )

    def __init__(self):
        super().__init__(validations=[self.VALIDATION_RDN_CONTAINS_MULTIPLE_ATVS],
                         pdu_class=rfc5280.RelativeDistinguishedName)

    def validate(self, node):
        child_count = len(node.children)

        if child_count > 1:
            raise validation.ValidationFindingEncountered(self.VALIDATION_RDN_CONTAINS_MULTIPLE_ATVS)


class InternalDomainNameValidator(validation.Validator):
    VALIDATION_INTERNAL_DOMAIN_NAME = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.internal_domain_name'
    )

    def __init__(self, *args, **kwargs):
        self._psl = publicsuffixlist.PublicSuffixList(accept_unknown=False)

        super().__init__(validations=[self.VALIDATION_INTERNAL_DOMAIN_NAME], **kwargs)

    def extract_domain_name(self, node):
        return str(node.pdu)

    def validate_with_value(self, node, value):
        if self._psl.publicsuffix(value) is None:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INTERNAL_DOMAIN_NAME,
                f'Internal domain name: "{value}"'
            )

    def validate(self, node):
        domain_name = self.extract_domain_name(node)

        return self.validate_with_value(node, domain_name)


class GeneralNameDnsNameInternalDomainNameValidator(InternalDomainNameValidator):
    def __init__(self, allow_onion_tld=False):
        self._allow_onion_tld = allow_onion_tld

        super().__init__(predicate=general_name.create_generalname_type_predicate('dNSName'))

    def validate_with_value(self, node, value):
        if len(value) == 0 and general_name.is_nameconstraints_child_node(node):
            return
        else:
            if self._allow_onion_tld and value.lower().endswith('.onion'):
                return
            else:
                return super().validate_with_value(node, value)


class UriInternalDomainNameValidator(InternalDomainNameValidator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def extract_domain_name(self, node):
        return urlparse(str(node.pdu)).netloc


class GeneralNameUriInternalDomainNameValidator(InternalDomainNameValidator):
    def __init__(self):
        super().__init__(predicate=general_name.create_generalname_type_predicate('uniformResourceIdentifier'))

    def extract_domain_name(self, node):
        if general_name.is_nameconstraints_child_node(node):
            return str(node.pdu).lstrip('.')
        else:
            return urlparse(str(node.pdu)).hostname

    def validate_with_value(self, node, value):
        if len(value) == 0 and general_name.is_nameconstraints_child_node(node):
            return
        else:
            return super().validate_with_value(node, value)


class EmailAddressInternalDomainNameValidator(InternalDomainNameValidator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def extract_domain_name(self, node):
        _, domain_part = str(node.pdu).split('@', maxsplit=1)

        return domain_part


class GeneralNameRfc822NameInternalDomainNameValidator(EmailAddressInternalDomainNameValidator):
    def __init__(self):
        super().__init__(predicate=general_name.create_generalname_type_predicate('rfc822Name'))

    def extract_domain_name(self, node):
        if general_name.is_nameconstraints_child_node(node):
            return str(node.pdu).lstrip('.')
        else:
            return super().extract_domain_name(node)

    def validate_with_value(self, node, value):
        if len(value) == 0 and general_name.is_nameconstraints_child_node(node):
            return
        else:
            super().validate_with_value(node, value)


class SmtpUtf8MailboxInternalDomainNameValidator(EmailAddressInternalDomainNameValidator):
    def __init__(self):
        super().__init__(pdu_class=rfc8398.SmtpUTF8Mailbox)

    def extract_domain_name(self, node):
        domain_part = super().extract_domain_name(node)

        return domain_part.encode('idna').decode()


class InternalIpAddressValidator(validation.Validator):
    VALIDATION_INTERNAL_IP_ADDRESS = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.internal_ip_address'
    )

    def __init__(self, *args, **kwargs):
        super().__init__(validations=self.VALIDATION_INTERNAL_IP_ADDRESS, **kwargs)

    @staticmethod
    def _extract_ip_address(node):
        octets = node.pdu.asOctets()

        if len(octets) == 4:
            return ipaddress.IPv4Address(octets)
        else:
            return ipaddress.IPv6Address(octets)

    def validate(self, node):
        ip_addr = InternalIpAddressValidator._extract_ip_address(node)

        if not ip_addr.is_global:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INTERNAL_IP_ADDRESS,
                f'Internal IP address: "{ip_addr}"'
            )


class GeneralNameInternalIpAddressValidator(InternalIpAddressValidator):
    def __init__(self):
        super().__init__(predicate=general_name.create_generalname_type_predicate('iPAddress'))

    def validate(self, node):
        if general_name.is_nameconstraints_child_node(node):
            return

        super().validate(node)


class OrganizationNameTruncatedLegalNameValidator(validation.Validator):
    VALIDATION_ORGANIZATION_NAME_TRUNCATED_LEGAL_NAME = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.organization_name_no_closing_parenthesis'
    )

    def __init__(self):
        super().__init__(validations=[self.VALIDATION_ORGANIZATION_NAME_TRUNCATED_LEGAL_NAME],
                         pdu_class=rfc5280.X520OrganizationName)

    def validate(self, node):
        _, value_node = node.child

        value = unicodedata.normalize('NFC', str(value_node.pdu))

        open_paren_idx = value.find('(')
        if open_paren_idx < 0:
            return

        close_paren_idx = value.find(')', open_paren_idx)

        if close_paren_idx < 0:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_ORGANIZATION_NAME_TRUNCATED_LEGAL_NAME,
                f'Organization name attribute with truncated legal name: "{value}"'
            )



import functools
import glob
import sys
from os import path

from pkilint.cabf import serverauth
from pkilint.cabf.serverauth import serverauth_constants
from pkilint.pkix import certificate
from tests import integration_certificate

this_module = sys.modules[__name__]


for certificate_type in serverauth_constants.CertificateType:
    cur_dir = path.dirname(__file__)

    test_dir = path.join(cur_dir, 'tls_br', certificate_type.name.lower())

    files = glob.glob(path.join(test_dir, '*.crttest'))

    for file in files:
        validator = certificate.create_pkix_certificate_validator_container(
            serverauth.create_decoding_validators(),
            serverauth.create_validators(certificate_type)
        )
        filters = serverauth.create_serverauth_finding_filters(certificate_type)

        file_no_ext, _ = path.splitext(path.basename(file))

        func_name = f'test_{certificate_type}_{file_no_ext}'

        setattr(this_module, func_name, functools.partial(integration_certificate.run_test, file, validator, filters))



import re

from pyasn1.type import univ, char, constraint

from pkilint import validation

ISO_OID_ARC = univ.ObjectIdentifier('1.3.6.1.4.1.52266')


id_ce_lei = ISO_OID_ARC + (1,)
id_ce_role = ISO_OID_ARC + (2,)

ub_leiRole_length = 100


class Lei(char.PrintableString):
    subtypeSpec = constraint.ValueSizeConstraint(20, 20)


class Role(char.PrintableString):
    subtypeSpec = constraint.ValueSizeConstraint(1, ub_leiRole_length)


EXTENSION_MAPPINGS = {
    id_ce_lei: Lei(),
    id_ce_role: Role(),
}


_LEI_FORMAT_RE = re.compile(r'^(?P<value>[0-9A-Z]{18})(?P<checksum>\d{2})$')


VALIDATION_INVALID_LEI_FORMAT = validation.ValidationFinding(
    validation.ValidationFindingSeverity.ERROR,
    'iso.lei.invalid_lei_format'
)

VALIDATION_INVALID_LEI_CHECKSUM = validation.ValidationFinding(
    validation.ValidationFindingSeverity.ERROR,
    'iso.lei.invalid_lei_checksum'
)


def _calculate_checksum(value):
    subs = {chr(l): str(10 + l - ord('A')) for l in range(ord('A'), ord('Z') + 1)}

    for l, n in subs.items():
        value = value.replace(l, n)

    value += '00'

    value_int = int(value)

    remainder = value_int % 97

    return 98 - remainder


def validate_lei(lei: str):
    m = _LEI_FORMAT_RE.match(lei)

    if m is None:
        raise validation.ValidationFindingEncountered(
            VALIDATION_INVALID_LEI_FORMAT, f'Invalid LEI format: {lei}'
        )

    value_part = m.group('value')
    actual_checksum = int(m.group('checksum'))

    expected_checksum = _calculate_checksum(value_part)

    if actual_checksum != expected_checksum:
        raise validation.ValidationFindingEncountered(
            VALIDATION_INVALID_LEI_CHECKSUM, f'Expected: {expected_checksum}, actual: {actual_checksum}'
        )


class LeiExtensionValueSyntaxValidator(validation.Validator):
    def __init__(self):
        super().__init__(validations=[VALIDATION_INVALID_LEI_FORMAT, VALIDATION_INVALID_LEI_CHECKSUM],
                         pdu_class=Lei)

    def validate(self, node):
        value = str(node.pdu)

        validate_lei(value)



from pyasn1_alt_modules import rfc5280

from pkilint import validation


class CorrectVersionValidator(validation.ScalarFieldValueEqualityValidator):
    def __init__(self):
        super().__init__(
            path='certificate.tbsCertificate.version',
            value=rfc5280.Version.namedValues['v3'],
            validations=validation.ValidationFinding(
                validation.ValidationFindingSeverity.ERROR,
                'pkix.certificate_version_is_not_v3'
            )
        )


class SignatureAlgorithmMatchValidator(validation.DEREqualityValidator):
    def __init__(self):
        super().__init__(
            other_node_retriever=(
                lambda n: n.navigate('^.tbsCertificate.signature')
            ),
            path='certificate.signatureAlgorithm',
            validation=validation.ValidationFinding(
                validation.ValidationFindingSeverity.ERROR,
                'pkix.certificate_signature_algorithm_mismatch'
            )
        )


class IssuerUniqueIdAbsenceValidator(validation.NodePresenceValidator):
    VALIDATION_ISSUER_UNIQUE_ID_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.issuer_unique_id_present'
    )

    def __init__(self):
        super().__init__(
            node_retriever=lambda n: n.navigate('issuerUniqueID'),
            presence_finding=self.VALIDATION_ISSUER_UNIQUE_ID_PRESENT,
            pdu_class=rfc5280.TBSCertificate
        )


class SubjectUniqueIdAbsenceValidator(validation.NodePresenceValidator):
    VALIDATION_SUBJECT_UNIQUE_ID_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.subject_unique_id_present'
    )

    def __init__(self):
        super().__init__(
            node_retriever=lambda n: n.navigate('subjectUniqueID'),
            presence_finding=self.VALIDATION_SUBJECT_UNIQUE_ID_PRESENT,
            pdu_class=rfc5280.TBSCertificate
        )



from fastapi import HTTPException

from pkilint.cabf import serverauth
from pkilint.cabf.serverauth import serverauth_constants
from pkilint.pkix import certificate
from pkilint.rest import model


class CabfServerauthLinterGroup(model.LinterGroup):
    def __init__(self, linters):
        super().__init__(name='cabf-serverauth', linters=linters)

    def determine_linter(self, doc):
        cert_type = serverauth.determine_certificate_type(doc)

        # this doesn't fail, so we don't need to guard against not being able to determine the certificate type
        return next((l for l in self.linters if l.name.casefold() == cert_type.to_option_str.casefold()))


def create_linter_group_instance():
    return CabfServerauthLinterGroup(
        [
            model.Linter(
                validator=certificate.create_pkix_certificate_validator_container(
                    serverauth.create_decoding_validators(),
                    serverauth.create_validators(cert_type)
                ),
                finding_filters=serverauth.create_serverauth_finding_filters(cert_type),
                name=cert_type.to_option_str
            )
            for cert_type in serverauth_constants.CertificateType
        ]
    )



from pkilint import validation, document
from pkilint.pkix import time


class OCSPSaneValidityPeriodValidator(time.SaneValidityPeriodValidator):
    VALIDATION_NEGATIVE_VALIDITY_PERIOD = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.ocsp_negative_validity_period'
    )

    def __init__(self):
        super().__init__(
            end_validity_node_retriever=lambda n: n.navigate('^.nextUpdate'),
            path_re=document.get_re_for_path_glob(
                'oCSPResponse.responseBytes.response.basicOCSPResponse.tbsResponseData.responses.*.thisUpdate'
            ),
            predicate=lambda n: 'nextUpdate' in n.parent.children,
            validation=self.VALIDATION_NEGATIVE_VALIDITY_PERIOD
        )



from pyasn1_alt_modules import rfc5280, rfc3279, rfc5480, rfc8398, rfc8410, rfc3739

import pkilint.adobe.asn1 as adobe_asn1
from pkilint import oid, validation
from pkilint.cabf.smime import smime_constants
from pkilint.cabf.smime.smime_constants import Generation, ValidationLevel
from pkilint.cabf.smime.smime_name import get_email_addresses_from_san
from pkilint.iso import lei
from pkilint.pkix import extension


class CertificatePoliciesPresenceValidator(extension.ExtensionPresenceValidator):
    VALIDATION_CERTIFICATE_POLICIES_EXTENSION_ABSENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.certificate_policies_extension_missing'
    )

    def __init__(self):
        super().__init__(
            extension_oid=rfc5280.id_ce_certificatePolicies,
            validation=self.VALIDATION_CERTIFICATE_POLICIES_EXTENSION_ABSENT,
            pdu_class=rfc5280.Extensions
        )


class ExtendedKeyUsagePresenceValidator(extension.ExtensionPresenceValidator):
    VALIDATION_EKU_EXTENSION_ABSENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.extended_key_usage_extension_missing'
    )

    def __init__(self):
        super().__init__(
            extension_oid=rfc5280.id_ce_extKeyUsage,
            validation=self.VALIDATION_EKU_EXTENSION_ABSENT,
            pdu_class=rfc5280.Extensions
        )


class CabfSmimeKeyUsagePresenceValidator(extension.ExtensionPresenceValidator):
    VALIDATION_KU_EXTENSION_ABSENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.key_usage_extension_missing'
    )

    def __init__(self):
        super().__init__(
            extension_oid=rfc5280.id_ce_keyUsage,
            validation=self.VALIDATION_KU_EXTENSION_ABSENT,
            pdu_class=rfc5280.Extensions
        )


class SubjectAlternativeNamePresenceValidator(
    extension.ExtensionPresenceValidator
):
    VALIDATION_SAN_EXTENSION_ABSENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.san_extension_missing'
    )

    def __init__(self):
        super().__init__(
            extension_oid=rfc5280.id_ce_subjectAltName,
            validation=self.VALIDATION_SAN_EXTENSION_ABSENT,
            pdu_class=rfc5280.Extensions
        )


class LeiPresenceValidator(validation.Validator):
    VALIDATION_LEI_EXTENSION_PROHIBITED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.lei_extension_prohibited'
    )

    def __init__(self, validation_level):
        super().__init__(validations=[self.VALIDATION_LEI_EXTENSION_PROHIBITED],
                         pdu_class=lei.Lei
                         )

        self._validation_level = validation_level

    def validate(self, node):
        if self._validation_level in {ValidationLevel.MAILBOX, ValidationLevel.INDIVIDUAL}:
            raise validation.ValidationFindingEncountered(self.VALIDATION_LEI_EXTENSION_PROHIBITED)


class LeiRolePresenceValidator(validation.Validator):
    VALIDATION_LEI_ROLE_EXTENSION_PROHIBITED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.lei_role_extension_prohibited'
    )

    def __init__(self, validation_level):
        super().__init__(validations=[self.VALIDATION_LEI_ROLE_EXTENSION_PROHIBITED],
                         pdu_class=lei.Role
                         )

        self._validation_level = validation_level

    def validate(self, node):
        if self._validation_level != ValidationLevel.SPONSORED:
            raise validation.ValidationFindingEncountered(self.VALIDATION_LEI_ROLE_EXTENSION_PROHIBITED)


class LeiCriticalityValidator(extension.ExtensionCriticalityValidator):
    VALIDATION_LEI_EXTENSION_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.lei_extension_critical'
    )

    def __init__(self):
        super().__init__(type_oid=lei.id_ce_lei, is_critical=False, validation=self.VALIDATION_LEI_EXTENSION_CRITICAL)


class LeiRoleCriticalityValidator(extension.ExtensionCriticalityValidator):
    VALIDATION_LEI_ROLE_EXTENSION_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.lei_role_extension_critical'
    )

    def __init__(self):
        super().__init__(type_oid=lei.id_ce_role, is_critical=False,
                         validation=self.VALIDATION_LEI_ROLE_EXTENSION_CRITICAL)


class RequiredPolicyIdentifierValidator(validation.Validator):
    VALIDATION_NO_CABF_RESERVED_OID = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.no_required_reserved_policy_oid'
    )

    VALIDATION_MULTIPLE_RESERVED_OIDS = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.multiple_reserved_policy_oids'
    )

    VALIDATION_ANYPOLICY_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.anypolicy_present'
    )

    def __init__(self, validation_level, generation):
        self._validation_level = validation_level
        self._generation = generation

        super().__init__(validations=[
            self.VALIDATION_NO_CABF_RESERVED_OID,
            self.VALIDATION_MULTIPLE_RESERVED_OIDS,
            self.VALIDATION_ANYPOLICY_PRESENT,
        ], pdu_class=rfc5280.CertificatePolicies)

    def validate(self, node):
        oids = set(
            (
                pi.children['policyIdentifier'].pdu
                for pi in node.children.values()
            )
        )

        findings = []

        if rfc5280.anyPolicy in oids:
            findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_ANYPOLICY_PRESENT,
                    None
                )
            )

        expected_oid = smime_constants.get_policy_oid(
            self._validation_level, self._generation
        )

        all_smime_oids = set()
        for g in smime_constants.Generation:
            for v in smime_constants.ValidationLevel:
                all_smime_oids.add(smime_constants.get_policy_oid(v, g))

        cert_smime_oids = oids.intersection(all_smime_oids)

        if expected_oid not in cert_smime_oids:
            findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_NO_CABF_RESERVED_OID,
                    f'Required policy OID {str(expected_oid)} is missing'
                )
            )

        if len(cert_smime_oids) > 1:
            desc = ', '.join(map(str, cert_smime_oids))

            findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_MULTIPLE_RESERVED_OIDS,
                    f'Multiple CA/B Reserved Policy OIDs found: {desc}'
                )
            )

        return validation.ValidationResult(self, node, findings)


class SmimeUserNoticeValidator(validation.Validator):
    VALIDATION_USERNOTICE_HAS_NOTICEREF = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.usernotice_has_noticeref'
    )

    def __init__(self):
        super().__init__(
            validations=self.VALIDATION_USERNOTICE_HAS_NOTICEREF,
            pdu_class=rfc5280.UserNotice
        )

    def validate(self, node):
        if 'noticeRef' in node.children:
            raise validation.ValidationFindingEncountered(self.VALIDATION_USERNOTICE_HAS_NOTICEREF)


class AllowedExtendedKeyUsageValidator(validation.Validator):
    VALIDATION_EMAIL_PROTECTION_EKU_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.emailprotection_eku_missing'
    )

    VALIDATION_PROHIBITED_EKU_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.prohibited_eku_present'
    )

    _LEGACY_MP_PROHIBITED_EKUS = {rfc5280.anyExtendedKeyUsage, rfc5280.id_kp_codeSigning, rfc5280.id_kp_OCSPSigning,
                                  rfc5280.id_kp_serverAuth, rfc5280.id_kp_timeStamping}

    def __init__(self, generation):
        self._generation = generation

        super().__init__(
            validations=[
                self.VALIDATION_EMAIL_PROTECTION_EKU_MISSING,
                self.VALIDATION_PROHIBITED_EKU_PRESENT,
            ],
            pdu_class=rfc5280.ExtKeyUsageSyntax
        )

    def validate(self, node):
        kp_oids = set(
            (
                kp.pdu
                for kp in node.children.values()
            )
        )

        findings = []
        if rfc5280.id_kp_emailProtection not in kp_oids:
            findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_EMAIL_PROTECTION_EKU_MISSING,
                    None
                )
            )

        if self._generation == smime_constants.Generation.STRICT:
            prohibited_kps = kp_oids.difference({rfc5280.id_kp_emailProtection})
        else:
            prohibited_kps = kp_oids.intersection(self._LEGACY_MP_PROHIBITED_EKUS)

        if len(prohibited_kps) > 0:
            findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_PROHIBITED_EKU_PRESENT,
                    f'Prohibited EKU(s) present: {oid.format_oids(prohibited_kps)}'
                )
            )

        return validation.ValidationResult(self, node, findings)


def _has_bit(node, bit_name):
    bit = rfc5280.KeyUsage.namedValues[bit_name]
    return len(node.pdu) > bit and node.pdu[bit] != 0


def _get_prohibited_kus(node, allowed_kus):
    prohibited_kus = (
        set(map(str, rfc5280.KeyUsage.namedValues)).difference(
            allowed_kus
        )
    )

    return [
        pk
        for pk in prohibited_kus
        if _has_bit(node, pk)
    ]


class AllowedKeyUsageValidator(validation.Validator):
    VALIDATION_UNKNOWN_CERT_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.unknown_certificate_key_usage_type'
    )

    VALIDATION_REQUIRED_KU_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.required_ku_missing'
    )

    VALIDATION_PROHIBITED_KU_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.prohibited_ku_present'
    )

    VALIDATION_UNSUPPORTED_PUBKEY_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.unsupported_public_key_type'
    )

    def __init__(self, generation):
        super().__init__(
            validations=[
                self.VALIDATION_REQUIRED_KU_MISSING,
                self.VALIDATION_PROHIBITED_KU_PRESENT,
                self.VALIDATION_UNKNOWN_CERT_TYPE,
                self.VALIDATION_UNSUPPORTED_PUBKEY_TYPE,
            ],
            pdu_class=rfc5280.KeyUsage
        )

        self._generation = generation

    def validate(self, node):
        spki_alg_oid = node.navigate(
            ':certificate.tbsCertificate.subjectPublicKeyInfo.algorithm.algorithm'
        ).pdu

        DS = 'digitalSignature'
        NR = 'nonRepudiation'
        DE = 'dataEncipherment'
        KE = 'keyEncipherment'
        KA = 'keyAgreement'
        EO = 'encipherOnly'
        DO = 'decipherOnly'

        allowed_kus = {DS}
        is_signing_cert = _has_bit(node, DS)

        if is_signing_cert:
            allowed_kus.add(NR)

        if spki_alg_oid == rfc3279.rsaEncryption:
            allowed_kus.add(KE)

            is_key_mgmt_cert = _has_bit(node, KE)

            if is_key_mgmt_cert and self._generation != Generation.STRICT:
                allowed_kus.add(DE)
        elif spki_alg_oid == rfc5480.id_ecPublicKey:
            is_key_mgmt_cert = _has_bit(node, KA)

            if is_key_mgmt_cert:
                allowed_kus.update({EO, DO, KA})
        elif spki_alg_oid in {rfc8410.id_Ed448, rfc8410.id_Ed25519}:
            is_key_mgmt_cert = False
        else:
            raise validation.ValidationFindingEncountered(self.VALIDATION_UNSUPPORTED_PUBKEY_TYPE)

        if not is_signing_cert and not is_key_mgmt_cert:
            raise validation.ValidationFindingEncountered(self.VALIDATION_UNKNOWN_CERT_TYPE)

        prohibited_kus = _get_prohibited_kus(node, allowed_kus)

        if len(prohibited_kus) > 0:
            ku_str = ', '.join(sorted(prohibited_kus))

            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PROHIBITED_KU_PRESENT, f'Prohibited KUs present: {ku_str}'
            )


class EndEntityValidator(validation.Validator):
    VALIDATION_SMIME_CERT_IS_CA = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.is_ca_certificate'
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_SMIME_CERT_IS_CA],
            pdu_class=rfc5280.BasicConstraints
        )

    def validate(self, node):
        if bool(node.children['cA'].pdu):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_SMIME_CERT_IS_CA
            )


class SubjectAlternativeNameContainsEmailAddressValidator(validation.Validator):
    VALIDATION_SAN_EMAIL_ADDRESS_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.san_does_not_contain_email_address'
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_SAN_EMAIL_ADDRESS_MISSING],
            pdu_class=rfc5280.SubjectAltName
        )

    def validate(self, node):
        if len(get_email_addresses_from_san(node.document)) == 0:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_SAN_EMAIL_ADDRESS_MISSING
            )


class SubjectAlternativeNameProhibitedGeneralNameTypesValidator(
    validation.Validator
):
    VALIDATION_PROHIBITED_GENERALNAME_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.prohibited_generalname_type_present'
    )

    VALIDATION_PROHIBITED_OTHERNAME_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.prohibited_othername_type_present'
    )

    _ALLOWED_TYPES = {'rfc822Name', 'directoryName', 'otherName'}

    def __init__(self, generation):
        super().__init__(
            validations=[self.VALIDATION_PROHIBITED_GENERALNAME_TYPE, self.VALIDATION_PROHIBITED_OTHERNAME_TYPE],
            pdu_class=rfc5280.SubjectAltName
        )

        self._generation = generation

    def validate(self, node):
        present_types = set((
            gn.child[0]
            for gn in node.children.values()
        ))

        prohibited_present_types = present_types - self._ALLOWED_TYPES

        if len(prohibited_present_types) > 0:
            type_str = ', '.join(prohibited_present_types)

            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PROHIBITED_GENERALNAME_TYPE,
                f'Prohibited GeneralName type(s) present: {type_str}'
            )

        for gn in node.children.values():
            name, val = gn.child

            if name == 'otherName':
                type_oid = val.children['type-id'].pdu

                if type_oid != rfc8398.id_on_SmtpUTF8Mailbox and self._generation == Generation.STRICT:
                    raise validation.ValidationFindingEncountered(
                        self.VALIDATION_PROHIBITED_OTHERNAME_TYPE,
                        f'Prohibited otherName present: {type_oid}'
                    )


class CabfSmimeKeyUsageCriticalityValidator(extension.ExtensionCriticalityValidator):
    VALIDATION_KU_NOT_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'cabf.smime.ku_extension_not_critical'
    )

    def __init__(self):
        super().__init__(
            type_oid=rfc5280.id_ce_keyUsage,
            is_critical=True,
            validation=self.VALIDATION_KU_NOT_CRITICAL
        )


class GmailAuthorityInfoAccessCaIssuersValidator(validation.Validator):
    VALIDATION_AIA_CA_ISSUERS_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        'googl.gmail.authority_info_access_ca_issuers_missing'
    )

    def __init__(self):
        super().__init__(
            pdu_class=rfc5280.AuthorityInfoAccessSyntax,
            validations=[self.VALIDATION_AIA_CA_ISSUERS_MISSING]
        )

    def validate(self, node):
        if not any((
                ad for ad in node.children.values() if ad.children['accessMethod'].pdu == rfc5280.id_ad_caIssuers
        )):
            raise validation.ValidationFindingEncountered(self.VALIDATION_AIA_CA_ISSUERS_MISSING)


class AllowedCrldpFullNameValidator(validation.Validator):
    VALIDATION_CRLDP_FULLNAME_PROHIBITED_GENERALNAME_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.crldp_fullname_prohibited_generalname_type'
    )

    VALIDATION_CRLDP_FULLNAME_PROHIBITED_URI_SCHEME = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.crldp_fullname_prohibited_uri_scheme'
    )

    def __init__(self, generation):
        super().__init__(
            validations=[self.VALIDATION_CRLDP_FULLNAME_PROHIBITED_URI_SCHEME,
                         self.VALIDATION_CRLDP_FULLNAME_PROHIBITED_GENERALNAME_TYPE],
            pdu_class=rfc5280.DistributionPointName,
            predicate=lambda n: 'fullName' in n.children
        )

        self._generation = generation

    def validate(self, node):
        uris = [str(gn.child[1].pdu)
                for gn in node.navigate('fullName').children.values() if 'uniformResourceIdentifier' in gn.children]

        if len(uris) != len(node.navigate('fullName').children):
            raise validation.ValidationFindingEncountered(self.VALIDATION_CRLDP_FULLNAME_PROHIBITED_GENERALNAME_TYPE)

        allowed_schemes = {'http', 'ldap', 'ftp'} if self._generation == Generation.LEGACY else {'http'}
        for u in uris:
            scheme, _ = u.split(':', maxsplit=1)

            if scheme.lower() not in allowed_schemes:
                raise validation.ValidationFindingEncountered(self.VALIDATION_CRLDP_FULLNAME_PROHIBITED_URI_SCHEME,
                                                              f'Prohibited URI scheme: {scheme}')


class AllowedAiaUriSchemeValidator(validation.Validator):
    VALIDATION_PROHIBITED_GENERALNAME_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.aia_prohibited_generalname_type'
    )

    VALIDATION_PROHIBITED_URI_SCHEME = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.aia_prohibited_uri_scheme'
    )

    def __init__(self, generation):
        super().__init__(
            validations=[
                self.VALIDATION_PROHIBITED_GENERALNAME_TYPE, self.VALIDATION_PROHIBITED_URI_SCHEME
            ],
            pdu_class=rfc5280.AccessDescription,
            predicate=lambda n: n.parent is not None and isinstance(n.parent, rfc5280.AuthorityInfoAccessSyntax)
        )

        self._generation = generation

    def validate(self, node):
        desc_node = node.children['accessLocation']
        name, value = desc_node.child

        if name != 'uniformResourceIdentifier':
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PROHIBITED_GENERALNAME_TYPE,
                f'Prohibited AIA GeneralName type: {name}'
            )

        scheme = str(value.pdu).split(':', maxsplit=1)[0].lower()

        allowed_schemes = {'http', 'ftp', 'ldap'} if self._generation == Generation.LEGACY else {'http'}
        if scheme not in allowed_schemes:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PROHIBITED_URI_SCHEME,
                f'Prohibited AIA URI scheme: {scheme}'
            )


class SubjectDirectoryAttributesPresenceValidator(validation.Validator):
    VALIDATION_SDA_EXTENSION_PROHIBITED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.subject_directory_attributes_extension_prohibited'
    )

    def __init__(self, validation_level, generation):
        super().__init__(validations=[self.VALIDATION_SDA_EXTENSION_PROHIBITED],
                         pdu_class=rfc5280.SubjectDirectoryAttributes
                         )

        self._validation_level = validation_level
        self._generation = generation

    def validate(self, node):
        if self._validation_level == ValidationLevel.MAILBOX or self._generation != Generation.LEGACY:
            raise validation.ValidationFindingEncountered(self.VALIDATION_SDA_EXTENSION_PROHIBITED)


class QCStatementsCriticalityValidator(extension.ExtensionCriticalityValidator):
    VALIDATION_QCS_EXTENSION_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.qc_statements_extension_critical'
    )

    def __init__(self):
        super().__init__(type_oid=rfc3739.id_pe_qcStatements, is_critical=False,
                         validation=self.VALIDATION_QCS_EXTENSION_CRITICAL)


class AdobeTimestampCriticalityValidator(extension.ExtensionCriticalityValidator):
    VALIDATION_ADOBE_TIMESTAMP_EXTENSION_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.adobe_timestamp_extension_critical'
    )

    def __init__(self):
        super().__init__(type_oid=adobe_asn1.id_adobe_timestamp, is_critical=False,
                         validation=self.VALIDATION_ADOBE_TIMESTAMP_EXTENSION_CRITICAL)


class AdobeArchiveRevInfoCriticalityValidator(extension.ExtensionCriticalityValidator):
    VALIDATION_ADOBE_ARCHIVE_REVINFO_EXTENSION_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.adobe_archive_revinfo_extension_critical'
    )

    def __init__(self):
        super().__init__(type_oid=adobe_asn1.id_adobe_archiverevinfo, is_critical=False,
                         validation=self.VALIDATION_ADOBE_ARCHIVE_REVINFO_EXTENSION_CRITICAL)


class AdobeTimestampPresenceValidator(validation.Validator):
    VALIDATION_ADOBE_TIMESTAMP_EXTENSION_PROHIBITED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.adobe_timestamp_extension_prohibited'
    )

    def __init__(self, generation):
        self._generation = generation

        super().__init__(validations=[self.VALIDATION_ADOBE_TIMESTAMP_EXTENSION_PROHIBITED],
                         pdu_class=adobe_asn1.AdobeTimestamp)

    def validate(self, node):
        if self._generation == smime_constants.Generation.STRICT:
            raise validation.ValidationFindingEncountered(self.VALIDATION_ADOBE_TIMESTAMP_EXTENSION_PROHIBITED)


class AdobeArchiveRevInfoPresenceValidator(validation.Validator):
    VALIDATION_ADOBE_ARCHIVE_REVINFO_EXTENSION_PROHIBITED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.adobe_archive_revinfo_extension_prohibited'
    )

    def __init__(self, generation):
        self._generation = generation

        super().__init__(validations=[self.VALIDATION_ADOBE_ARCHIVE_REVINFO_EXTENSION_PROHIBITED],
                         pdu_class=adobe_asn1.AdobeArchiveRevInfo)

    def validate(self, node):
        if self._generation == smime_constants.Generation.STRICT:
            raise validation.ValidationFindingEncountered(self.VALIDATION_ADOBE_ARCHIVE_REVINFO_EXTENSION_PROHIBITED)


class CrlDistributionPointPresenceValidator(extension.ExtensionPresenceValidator):
    VALIDATION_CRLDP_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'cabf.smime.crldp_extension_missing'
    )

    def __init__(self):
        super().__init__(extension_oid=rfc5280.id_ce_cRLDistributionPoints,
                         validation=self.VALIDATION_CRLDP_MISSING,
                         pdu_class=rfc5280.Extensions
                         )



from pyasn1_alt_modules import rfc5280, rfc6960, rfc6962, rfc4262

from pkilint import validation

EXTENSION_MAPPINGS = {
    **rfc4262._certificateExtensionsMap,
    **rfc6962._certificateExtensionsMapUpdate,
    **rfc6960._certificateExtensionsMapUpdate,
    **rfc5280.certificateExtensionsMap
}


def get_criticality_from_decoded_node(node):
    ext_node = node.navigate('^.^')

    return bool(ext_node.children['critical'].pdu)


class PermittedExtensionValidator(validation.Validator):
    VALIDATION_CODE = 'pkix.unknown_extension'

    def __init__(self, *, known_oids,
                 severity=validation.ValidationFindingSeverity.ERROR):
        self.known_oids = known_oids

        finding = [validation.ValidationFinding(severity, self.VALIDATION_CODE)]

        super().__init__(validations=finding,
                         pdu_class=rfc5280.Extension
                         )

    def validate(self, node):
        type_node = node.children['extnID']
        oid = type_node.pdu

        if oid not in self.known_oids:
            raise validation.ValidationFindingEncountered(
                self.validations[0],
                f'Unknown extension type: {str(oid)}'
            )


class UniqueExtensionValidator(validation.Validator):
    VALIDATION_EXTENSION_NOT_UNIQUE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.duplicate_extension'
    )

    def __init__(self):
        super().__init__(validations=self.VALIDATION_EXTENSION_NOT_UNIQUE,
                         pdu_class=rfc5280.Extensions
                         )

    def validate(self, node):
        oids = set()
        for child in node.children.values():
            oid = child.children['extnID'].pdu
            if oid in oids:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_EXTENSION_NOT_UNIQUE,
                    f'Multiple extensions of type "{str(oid)}"'
                )

            oids.add(oid)


class ExtensionCriticalityValidator(validation.Validator):
    def __init__(self, type_oid, is_critical, validation):
        self._type_oid = type_oid
        self._is_critical = is_critical
        self._validation = validation

        super().__init__(pdu_class=rfc5280.Extension, validations=[validation])

    def match(self, node):
        if not super().match(node):
            return False

        node_oid = node.children['extnID'].pdu

        return self._type_oid == node_oid

    def validate(self, node):
        criticality = bool(node.navigate('critical').pdu)

        if self._is_critical and not criticality:
            raise validation.ValidationFindingEncountered(
                self._validation,
                f'Extension {self._type_oid} is not critical'
            )
        elif not self._is_critical and criticality:
            raise validation.ValidationFindingEncountered(
                self._validation,
                f'Extension {self._type_oid} is critical'
            )


class ExtensionsDecodingValidator(validation.DecodingValidator):
    def __init__(self, *, decode_func, **kwargs):
        super().__init__(decode_func=decode_func,
                         pdu_class=rfc5280.Extension,
                         **kwargs
                         )


class IssuerSubjectKeyIdentifierBinaryEqualValidator(validation.Validator):
    VALIDATION_AUTH_KEY_ID_EXTENSION_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.authority_key_id_extension_missing'
    )

    VALIDATION_AUTH_KEY_ID_KEYID_FIELD_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.authority_key_id_keyid_field_missing'
    )

    VALIDATION_KEY_IDENTIFIER_NOT_MATCH = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.issuer_and_subject_key_identifier_not_equal'
    )

    def __init__(self, *, subject_auth_key_id_retriever):
        self._subject_auth_key_id_retriever = subject_auth_key_id_retriever

        super().__init__(validations=[
            self.VALIDATION_AUTH_KEY_ID_EXTENSION_MISSING,
            self.VALIDATION_AUTH_KEY_ID_KEYID_FIELD_MISSING,
            self.VALIDATION_KEY_IDENTIFIER_NOT_MATCH
        ],
            pdu_class=rfc5280.SubjectKeyIdentifier
        )

    def validate(self, node):
        auth_id_ext_and_idx = self._subject_auth_key_id_retriever(node)

        if auth_id_ext_and_idx is None:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_AUTH_KEY_ID_EXTENSION_MISSING
            )

        ext, _ = auth_id_ext_and_idx

        decoded_auth_key_id = ext.navigate('extnValue.authorityKeyIdentifier')

        key_id_node = decoded_auth_key_id.children.get('keyIdentifier')
        if key_id_node is None:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_AUTH_KEY_ID_KEYID_FIELD_MISSING
            )

        if bytes(key_id_node.pdu) != bytes(node.pdu):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_KEY_IDENTIFIER_NOT_MATCH
            )


class AuthorityKeyIdentifierCriticalityValidator(ExtensionCriticalityValidator):
    VALIDATION_CRITICAL_AKI = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.authority_key_identifier_critical'
    )

    def __init__(self):
        super().__init__(type_oid=rfc5280.id_ce_authorityKeyIdentifier,
                         is_critical=False, validation=self.VALIDATION_CRITICAL_AKI
                         )


class ExtensionPresenceValidator(validation.Validator):
    def __init__(self, *, extension_oid, validation, **kwargs):
        self._extension_oid = extension_oid
        self._validation = validation

        super().__init__(validations=[validation], **kwargs)

    def validate(self, node):
        ext_idx = node.document.get_extension_by_oid(self._extension_oid)

        if ext_idx is None:
            raise validation.ValidationFindingEncountered(self._validation)


class AuthorityKeyIdentifierValidator(validation.Validator):
    VALIDATION_AKI_NO_KEY_ID = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.authority_key_identifier_keyid_missing'
    )

    VALIDATION_AKI_NO_SERIAL_NUMBER = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.aki_with_cert_issuer_but_serial_number_absent'
    )

    VALIDATION_AKI_NO_CERT_ISSUER = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.aki_with_serial_number_but_cert_issuer_absent'
    )

    def __init__(self):
        super().__init__(
            pdu_class=rfc5280.AuthorityKeyIdentifier,
            validations=[
                self.VALIDATION_AKI_NO_KEY_ID,
                self.VALIDATION_AKI_NO_CERT_ISSUER,
                self.VALIDATION_AKI_NO_SERIAL_NUMBER
            ]
        )

    def validate(self, node):
        results = []

        if 'keyIdentifier' not in node.children:
            results.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_AKI_NO_KEY_ID, None
                )
            )

        if 'authorityCertIssuer' in node.children and (
                'authorityCertSerialNumber' not in node.children
        ):
            results.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_AKI_NO_SERIAL_NUMBER, None
                )
            )

        if 'authorityCertSerialNumber' in node.children and (
                'authorityCertIssuer' not in node.children
        ):
            results.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_AKI_NO_CERT_ISSUER, None
                )
            )

        return validation.ValidationResult(self, node, results)


class DistributionPointValidator(validation.Validator):
    VALIDATION_DP_NO_NAME_OR_ISSUER = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.distribution_point_does_not_contain_name_or_issuer'
    )

    def __init__(self):
        super().__init__(validations=[self.VALIDATION_DP_NO_NAME_OR_ISSUER], pdu_class=rfc5280.DistributionPoint)

    def validate(self, node):
        if 'distributionPoint' not in node.children and 'cRLIssuer' not in node.children:
            raise validation.ValidationFindingEncountered(self.VALIDATION_DP_NO_NAME_OR_ISSUER)
